import hashlib
import json
import os
import threading
from typing import Optional, List, Callable

import paho.mqtt.client as mqtt

from platypush.backend import Backend
from platypush.config import Config
from platypush.context import get_plugin
from platypush.message import Message
from platypush.message.event.mqtt import MQTTMessageEvent
from platypush.message.request import Request
from platypush.plugins.mqtt import MqttPlugin as MQTTPlugin
from platypush.utils import set_thread_name


class MqttClient(mqtt.Client, threading.Thread):
    def __init__(self, *args, host: str, port: int, topics: Optional[List[str]] = None,
                 on_message: Optional[Callable] = None, username: Optional[str] = None, password: Optional[str] = None,
                 client_id: Optional[str] = None, tls_cafile: Optional[str] = None, tls_certfile: Optional[str] = None,
                 tls_keyfile: Optional[str] = None, tls_version: Optional = None, tls_ciphers: Optional = None,
                 tls_insecure: bool = False, keepalive: Optional[int] = 60, **kwargs):
        mqtt.Client.__init__(self, *args, client_id=client_id, **kwargs)
        threading.Thread.__init__(self)

        self.host = host
        self.port = port
        self.topics = set(topics or [])
        self.keepalive = keepalive
        self.on_connect = self.connect_hndl()

        if on_message:
            self.on_message = on_message

        if username and password:
            self.username_pw_set(username, password)

        if tls_cafile:
            self.tls_set(ca_certs=tls_cafile, certfile=tls_certfile, keyfile=tls_keyfile, tls_version=tls_version,
                         ciphers=tls_ciphers)

            self.tls_insecure_set(tls_insecure)

        self._running = False
        self._stop_scheduled = False

    def subscribe(self, *topics, **kwargs):
        """
        Client subscription handler.
        """
        if not topics:
            topics = self.topics

        self.topics.update(topics)
        for topic in topics:
            super().subscribe(topic, **kwargs)

    def unsubscribe(self, *topics, **kwargs):
        """
        Client unsubscription handler.
        """
        if not topics:
            topics = self.topics

        for topic in topics:
            super().unsubscribe(topic, **kwargs)
            self.topics.remove(topic)

    def connect_hndl(self):
        def handler(*_, **__):
            self.subscribe()

        return handler

    def run(self):
        super().run()
        self.connect(host=self.host, port=self.port, keepalive=self.keepalive)
        self._running = True
        self.loop_forever()

    def stop(self):
        if not self.is_alive():
            return

        self._stop_scheduled = True
        self.disconnect()
        self._running = False


class MqttBackend(Backend):
    """
    Backend that reads messages from a configured MQTT topic (default:
    ``platypush_bus_mq/<device_id>``) and posts them to the application bus.

    Triggers:

        * :class:`platypush.message.event.mqtt.MQTTMessageEvent` when a new
            message is received on one of the custom listeners

    Requires:

        * **paho-mqtt** (``pip install paho-mqtt``)
    """

    _default_mqtt_port = 1883

    def __init__(self, host: Optional[str] = None, port: int = _default_mqtt_port,
                 topic='platypush_bus_mq', subscribe_default_topic: bool = True,
                 tls_cafile: Optional[str] = None, tls_certfile: Optional[str] = None,
                 tls_keyfile: Optional[str] = None, tls_version: Optional[str] = None,
                 tls_ciphers: Optional[str] = None, tls_insecure: bool = False,
                 username: Optional[str] = None, password: Optional[str] = None,
                 client_id: Optional[str] = None, listeners=None,
                 *args, **kwargs):
        """
        :param host: MQTT broker host. If no host configuration is specified then
            the backend will use the host configuration specified on the ``mqtt``
            plugin if it's available.
        :param port: MQTT broker port (default: 1883)
        :param topic: Topic to read messages from (default: ``platypush_bus_mq/<device_id>``)
        :param subscribe_default_topic: Whether the backend should subscribe the default topic (default:
            ``platypush_bus_mq/<device_id>``) and execute the messages received there as action requests
            (default: True).
        :param tls_cafile: If TLS/SSL is enabled on the MQTT server and the certificate requires a certificate authority
            to authenticate it, `ssl_cafile` will point to the provided ca.crt file (default: None)
        :param tls_certfile: If TLS/SSL is enabled on the MQTT server and a client certificate it required, specify it
            here (default: None)
        :param tls_keyfile: If TLS/SSL is enabled on the MQTT server and a client certificate key it required,
            specify it here (default: None) :type tls_keyfile: str
        :param tls_version: If TLS/SSL is enabled on the MQTT server and it requires a certain TLS version, specify it
            here (default: None). Supported versions: ``tls`` (automatic), ``tlsv1``, ``tlsv1.1``, ``tlsv1.2``.
        :param tls_ciphers: If TLS/SSL is enabled on the MQTT server and an explicit list of supported ciphers is
            required, specify it here (default: None)
        :param tls_insecure: Set to True to ignore TLS insecure warnings (default: False).
        :param username: Specify it if the MQTT server requires authentication (default: None)
        :param password: Specify it if the MQTT server requires authentication (default: None)
        :param client_id: ID used to identify the client on the MQTT server (default: None).
            If None is specified then ``Config.get('device_id')`` will be used.
        :param listeners: If specified then the MQTT backend will also listen for
            messages on the additional configured message queues. This parameter
            is a list of maps where each item supports the same arguments passed
            to the main backend configuration (host, port, topic, password etc.).
            Note that the message queue configured on the main configuration
            will expect valid Platypush messages that then can execute, while
            message queues registered to the listeners will accept any message. Example::

                listeners:
                    - host: localhost
                      topics:
                          - topic1
                          - topic2
                          - topic3
                    - host: sensors
                      topics:
                          - topic4
                          - topic5

        """

        super().__init__(*args, **kwargs)

        if host:
            self.host = host
            self.port = port
            self.tls_cafile = self._expandpath(tls_cafile) if tls_cafile else None
            self.tls_certfile = self._expandpath(tls_certfile) if tls_certfile else None
            self.tls_keyfile = self._expandpath(tls_keyfile) if tls_keyfile else None
            self.tls_version = MQTTPlugin.get_tls_version(tls_version)
            self.tls_ciphers = tls_ciphers
            self.tls_insecure = tls_insecure
            self.username = username
            self.password = password
            self.client_id = client_id or Config.get('device_id')
        else:
            client = get_plugin('mqtt')
            assert client.host, 'No host specified on backend.mqtt nor mqtt configuration'

            self.host = client.host
            self.port = client.port
            self.tls_cafile = client.tls_cafile
            self.tls_certfile = client.tls_certfile
            self.tls_keyfile = client.tls_keyfile
            self.tls_version = client.tls_version
            self.tls_ciphers = client.tls_ciphers
            self.tls_insecure = client.tls_insecure
            self.username = client.username
            self.password = client.password
            self.client_id = client_id or client.client_id

        self.topic = '{}/{}'.format(topic, self.device_id)
        self.subscribe_default_topic = subscribe_default_topic
        self._listeners = {}  # client_id -> MqttClient map
        self.listeners_conf = listeners or []

    def send_message(self, msg, topic: Optional[str] = None, **kwargs):
        try:
            client = get_plugin('mqtt')
            client.send_message(topic=topic or self.topic, msg=msg, host=self.host,
                                port=self.port, username=self.username,
                                password=self.password, tls_cafile=self.tls_cafile,
                                tls_certfile=self.tls_certfile, tls_keyfile=self.tls_keyfile,
                                tls_version=self.tls_version, tls_insecure=self.tls_insecure,
                                tls_ciphers=self.tls_ciphers, **kwargs)
        except Exception as e:
            self.logger.exception(e)

    @staticmethod
    def _expandpath(path: str) -> str:
        return os.path.abspath(os.path.expanduser(path)) if path else path

    def add_listeners(self, *listeners):
        # noinspection PyShadowingNames,PyUnusedLocal
        for i, listener in enumerate(listeners):
            host = listener.get('host')
            if host:
                port = listener.get('port', self._default_mqtt_port)
                username = listener.get('username')
                password = listener.get('password')
                tls_cafile = self._expandpath(listener.get('tls_cafile'))
                tls_certfile = self._expandpath(listener.get('tls_certfile'))
                tls_keyfile = self._expandpath(listener.get('tls_keyfile'))
                tls_version = MQTTPlugin.get_tls_version(listener.get('tls_version'))
                tls_ciphers = listener.get('tls_ciphers')
                tls_insecure = listener.get('tls_insecure')
            else:
                host = self.host
                port = self.port
                username = self.username
                password = self.password
                tls_cafile = self.tls_cafile
                tls_certfile = self.tls_certfile
                tls_keyfile = self.tls_keyfile
                tls_version = self.tls_version
                tls_ciphers = self.tls_ciphers
                tls_insecure = self.tls_insecure

            topics = listener.get('topics')
            if not topics:
                self.logger.warning('No list of topics specified for listener n.{}'.format(i+1))
                continue

            client = self._get_client(host=host, port=port, topics=topics, username=username, password=password,
                                      client_id=self.client_id, tls_cafile=tls_cafile, tls_certfile=tls_certfile,
                                      tls_keyfile=tls_keyfile, tls_version=tls_version, tls_ciphers=tls_ciphers,
                                      tls_insecure=tls_insecure)

            if not client.is_alive():
                client.start()

    def _get_client_id(
            self, host: str, port: int, topics: Optional[List[str]] = None,
            client_id: Optional[str] = None, on_message: Optional[bool] = None,
    ) -> str:
        return '{client_id}-{client_hash}'.format(
            client_id=client_id or self.client_id,
            client_hash=hashlib.sha1('|'.join([
                host, str(port),
                json.dumps(sorted(topics or [])),
                str(id(on_message))
            ]).encode()).hexdigest(),
        )

    def _get_client(self, host: str, port: int, topics: Optional[List[str]] = None, username: Optional[str] = None,
                    password: Optional[str] = None, client_id: Optional[str] = None, tls_cafile: Optional[str] = None,
                    tls_certfile: Optional[str] = None, tls_keyfile: Optional[str] = None, tls_version: Optional = None,
                    tls_ciphers: Optional = None, tls_insecure: bool = False, on_message: Optional[Callable] = None) \
            -> MqttClient:
        on_message = on_message or self.on_mqtt_message()
        client_id = self._get_client_id(host=host, port=port, topics=topics, client_id=client_id, on_message=on_message)
        client = self._listeners.get(client_id)

        if not (client and client.is_alive()):
            client = self._listeners[client_id] = MqttClient(
                host=host, port=port, topics=topics, username=username, password=password,
                client_id=client_id, tls_cafile=tls_cafile, tls_certfile=tls_certfile,
                tls_keyfile=tls_keyfile, tls_version=tls_version, tls_ciphers=tls_ciphers,
                tls_insecure=tls_insecure, on_message=on_message
            )

        client.subscribe(*topics)
        return client

    def on_mqtt_message(self):
        def handler(client, __, msg):
            data = msg.payload
            # noinspection PyBroadException
            try:
                data = data.decode('utf-8')
                data = json.loads(data)
            except Exception as e:
                self.logger.debug(str(e))

            # noinspection PyProtectedMember
            self.bus.post(MQTTMessageEvent(host=client._host, port=client._port, topic=msg.topic, msg=data))

        return handler

    def on_exec_message(self):
        def handler(_, __, msg):
            # noinspection PyShadowingNames
            def response_thread(msg):
                set_thread_name('MQTTProcessor')
                response = self.get_message_response(msg)
                if not response:
                    return
                response_topic = '{}/responses/{}'.format(self.topic, msg.id)

                self.logger.info('Processing response on the MQTT topic {}: {}'.
                                 format(response_topic, response))

                self.send_message(response, topic=response_topic)

            msg = msg.payload.decode('utf-8')
            # noinspection PyBroadException
            try:
                msg = json.loads(msg)
                msg = Message.build(msg)
            except Exception as e:
                self.logger.debug(str(e))

            if not msg:
                return

            self.logger.info('Received message on the MQTT backend: {}'.format(msg))

            try:
                self.on_message(msg)
            except Exception as e:
                self.logger.exception(e)
                return

            if isinstance(msg, Request):
                threading.Thread(target=response_thread, name='MQTTProcessor', args=(msg,)).start()

        return handler

    def run(self):
        super().run()

        if self.host and self.subscribe_default_topic:
            topics = [self.topic]
            client = self._get_client(host=self.host, port=self.port, topics=topics, username=self.username,
                                      password=self.password, client_id=self.client_id,
                                      tls_cafile=self.tls_cafile, tls_certfile=self.tls_certfile,
                                      tls_keyfile=self.tls_keyfile, tls_version=self.tls_version,
                                      tls_ciphers=self.tls_ciphers, tls_insecure=self.tls_insecure,
                                      on_message=self.on_exec_message())

            client.start()
            self.logger.info('Initialized MQTT backend on host {}:{}, topic {}'.
                             format(self.host, self.port, self.topic))

        self.add_listeners(*self.listeners_conf)

    def on_stop(self):
        self.logger.info('Received STOP event on the MQTT backend')

        for listener in self._listeners.values():
            try:
                listener.stop()
            except Exception as e:
                self.logger.warning(f'Could not stop MQTT listener: {e}')

        self.logger.info('MQTT backend terminated')

# vim:sw=4:ts=4:et:



from enum import Enum

from platypush.plugins.camera.model.writer.ffmpeg import MKVStreamWriter, H264StreamWriter, H265StreamWriter
from platypush.plugins.camera.model.writer.image import JPEGStreamWriter, PNGStreamWriter, BMPStreamWriter, \
    MJPEGStreamWriter


class StreamHandlers(Enum):
    JPG = JPEGStreamWriter
    JPEG = JPEGStreamWriter
    PNG = PNGStreamWriter
    BMP = BMPStreamWriter
    MJPEG = MJPEGStreamWriter
    MJPG = MJPEGStreamWriter
    MKV = MKVStreamWriter
    WEBM = MKVStreamWriter
    H264 = H264StreamWriter
    H265 = H265StreamWriter
    MP4 = H264StreamWriter


# vim:sw=4:ts=4:et:



from platypush.message.response import Response


class TranslateResponse(Response):
    def __init__(self,
                 translated_text: str,
                 source_text: str,
                 detected_source_language: str,
                 *args,
                 **kwargs):
        super().__init__(*args, output={
            'translated_text': translated_text,
            'source_text': source_text,
            'detected_source_language': detected_source_language,
        }, **kwargs)


# vim:sw=4:ts=4:et:



import signal
import subprocess
from typing import Optional, Tuple

from PIL import Image
from PIL.Image import Image as ImageType

from platypush.plugins.camera import CameraPlugin
from platypush.plugins.camera.ffmpeg.model import FFmpegCamera, FFmpegCameraInfo


class CameraFfmpegPlugin(CameraPlugin):
    """
    Plugin to interact with a camera over FFmpeg.

    Requires:

        * **ffmpeg** package installed on the system.

    """

    _camera_class = FFmpegCamera
    _camera_info_class = FFmpegCameraInfo

    def __init__(self, device: Optional[str] = '/dev/video0', input_format: str = 'v4l2', ffmpeg_args: Tuple[str] = (),
                 **opts):
        """
        :param device: Path to the camera device (default: ``/dev/video0``).
        :param input_format: FFmpeg input format for the the camera device (default: ``v4l2``).
        :param ffmpeg_args: Extra options to be passed to the FFmpeg executable.
        :param opts: Camera options - see constructor of :class:`platypush.plugins.camera.CameraPlugin`.
        """
        super().__init__(device=device, input_format=input_format, **opts)
        self.camera_info.ffmpeg_args = ffmpeg_args or ()

    def prepare_device(self, camera: FFmpegCamera) -> subprocess.Popen:
        warmup_seconds = self._get_warmup_seconds(camera)
        ffmpeg = [camera.info.ffmpeg_bin, '-y', '-f', camera.info.input_format, '-i', camera.info.device, '-s',
                  '{}x{}'.format(*camera.info.resolution), '-ss', str(warmup_seconds),
                  *(('-r', str(camera.info.fps)) if camera.info.fps else ()),
                  '-pix_fmt', 'rgb24', '-f', 'rawvideo', *camera.info.ffmpeg_args, '-']

        self.logger.info('Running FFmpeg: {}'.format(' '.join(ffmpeg)))
        proc = subprocess.Popen(ffmpeg, stdout=subprocess.PIPE)
        # Start in suspended mode
        proc.send_signal(signal.SIGSTOP)
        return proc

    def start_camera(self, camera: FFmpegCamera, preview: bool = False, *args, **kwargs):
        super().start_camera(*args, camera=camera, preview=preview, **kwargs)
        if camera.object:
            camera.object.send_signal(signal.SIGCONT)

    def release_device(self, camera: FFmpegCamera):
        if camera.object:
            camera.object.terminate()
            if camera.object.stdout:
                camera.object.stdout.close()
            camera.object = None

    def wait_capture(self, camera: FFmpegCamera) -> None:
        if camera.object and camera.object.poll() is None:
            try:
                camera.object.wait(timeout=camera.info.capture_timeout)
            except Exception as e:
                self.logger.warning('Error on FFmpeg capture wait: {}'.format(str(e)))

    def capture_frame(self, camera: FFmpegCamera, *args, **kwargs) -> Optional[ImageType]:
        raw_size = camera.info.resolution[0] * camera.info.resolution[1] * 3
        data = camera.object.stdout.read(raw_size)
        if len(data) < raw_size:
            return
        return Image.frombytes('RGB', camera.info.resolution, data)


# vim:sw=4:ts=4:et:



import datetime
import json
import logging

from flask import Blueprint, request, abort, jsonify

from platypush.exceptions.user import UserException
from platypush.user import UserManager

auth = Blueprint('auth', __name__)
log = logging.getLogger(__name__)

# Declare routes list
__routes__ = [
    auth,
]


@auth.route('/auth', methods=['POST'])
def auth_endpoint():
    """
    Authentication endpoint. It validates the user credentials provided over a JSON payload with the following
    structure:

        .. code-block:: json

            {
                "username": "USERNAME",
                "password": "PASSWORD",
                "expiry_days": "The generated token should be valid for these many days"
            }

    ``expiry_days`` is optional, and if omitted or set to zero the token will be valid indefinitely.

    Upon successful validation, a new JWT token will be generated using the service's self-generated RSA key-pair and it
    will be returned to the user. The token can then be used to authenticate API calls to ``/execute`` by setting the
    ``Authorization: Bearer <TOKEN_HERE>`` header upon HTTP calls.

    :return: Return structure:

        .. code-block:: json

            {
                "token": "<generated token here>"
            }
    """
    try:
        payload = json.loads(request.get_data(as_text=True))
        username, password = payload['username'], payload['password']
    except Exception as e:
        log.warning('Invalid payload passed to the auth endpoint: ' + str(e))
        abort(400)

    expiry_days = payload.get('expiry_days')
    expires_at = None
    if expiry_days:
        expires_at = datetime.datetime.now() + datetime.timedelta(days=expiry_days)

    user_manager = UserManager()

    try:
        return jsonify({
            'token': user_manager.generate_jwt_token(username=username, password=password, expires_at=expires_at),
        })
    except UserException as e:
        abort(401, str(e))



import logging
import os

from flask import Flask

from platypush import Config
from platypush.backend.http.app.utils import get_routes


def initialize_logger():
    logger = logging.getLogger('werkzeug')
    try:
        log_conf = Config.get('logging')
        if 'level' in log_conf:
            logger.setLevel(log_conf['level'])
    except Exception as e:
        logger.warning('Could not read logging level')
        logger.exception(e)


## Webapp initialization

initialize_logger()

base_folder = os.path.abspath(os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '..'))

template_folder = os.path.join(base_folder, 'webapp/dist')
static_folder = os.path.join(base_folder, 'webapp/dist/static')

application = Flask('platypush',
                    template_folder=template_folder,
                    static_folder=static_folder)

for route in get_routes():
    application.register_blueprint(route)


# vim:sw=4:ts=4:et:



from marshmallow import fields
from marshmallow.schema import Schema


class SwitchStatusSchema(Schema):
    id = fields.Raw(metadata=dict(description='Device unique ID'))
    name = fields.String(required=True, metadata=dict(description='Device name'))
    on = fields.Boolean(required=True, metadata=dict(description='True if the device is on, False otherwise'))



import subprocess
import threading
import time

from abc import ABC
from typing import Optional, Tuple

from PIL.Image import Image

from platypush.plugins.camera.model.camera import Camera
from platypush.plugins.camera.model.writer import VideoWriter, FileVideoWriter, StreamWriter


class FFmpegWriter(VideoWriter, ABC):
    """
    Generic FFmpeg encoder for camera frames.
    """

    def __init__(self, *args, input_file: str = '-', input_format: str = 'rawvideo', output_file: str = '-',
                 output_format: Optional[str] = None, pix_fmt: Optional[str] = None,
                 output_opts: Optional[Tuple] = None, **kwargs):
        super().__init__(*args, **kwargs)

        self.input_file = input_file
        self.input_format = input_format
        self.output_format = output_format
        self.output_file = output_file
        self.width, self.height = self.camera.effective_resolution()
        self.pix_fmt = pix_fmt
        self.output_opts = output_opts or ()

        self.logger.info('Starting FFmpeg. Command: {}'.format(' '.join(self.ffmpeg_args)))
        self.ffmpeg = subprocess.Popen(self.ffmpeg_args, stdin=subprocess.PIPE, stdout=subprocess.PIPE)

    @property
    def ffmpeg_args(self):
        return [self.camera.info.ffmpeg_bin, '-y',
                '-f', self.input_format,
                *(('-pix_fmt', self.pix_fmt) if self.pix_fmt else ()),
                '-s', '{}x{}'.format(self.width, self.height),
                '-r', str(self.camera.info.fps),
                '-i', self.input_file,
                *(('-f', self.output_format) if self.output_format else ()),
                *self.output_opts,
                *(('-vcodec', self.camera.info.output_codec) if self.camera.info.output_codec else ()),
                self.output_file]

    def is_closed(self):
        return self.closed or not self.ffmpeg or self.ffmpeg.poll() is not None

    def write(self, image: Image):
        if self.is_closed():
            return

        try:
            self.ffmpeg.stdin.write(image.convert('RGB').tobytes())
        except Exception as e:
            self.logger.warning('FFmpeg send error: {}'.format(str(e)))
            self.close()

    def close(self):
        if not self.is_closed():
            if self.ffmpeg and self.ffmpeg.stdin:
                try:
                    self.ffmpeg.stdin.close()
                except (IOError, OSError):
                    pass

            if self.ffmpeg:
                self.ffmpeg.terminate()
                try:
                    self.ffmpeg.wait(timeout=5.0)
                except subprocess.TimeoutExpired:
                    self.logger.warning('FFmpeg has not returned - killing it')
                    self.ffmpeg.kill()

            if self.ffmpeg and self.ffmpeg.stdout:
                try:
                    self.ffmpeg.stdout.close()
                except (IOError, OSError):
                    pass

        self.ffmpeg = None
        super().close()


class FFmpegFileWriter(FileVideoWriter, FFmpegWriter):
    """
    Write camera frames to a file using FFmpeg.
    """

    def __init__(self, *args, output_file: str, **kwargs):
        super().__init__(*args, output_file=output_file, pix_fmt='rgb24', **kwargs)


class FFmpegStreamWriter(StreamWriter, FFmpegWriter, ABC):
    """
    Stream camera frames using FFmpeg.
    """

    def __init__(self, *args, output_format: str, output_opts: Optional[Tuple] = None, **kwargs):
        super().__init__(*args, pix_fmt='rgb24', output_format=output_format, output_opts=output_opts or (
            '-tune', 'zerolatency', '-preset', 'superfast', '-trellis', '0',
            '-fflags', 'nobuffer'), **kwargs)
        self._reader = threading.Thread(target=self._reader_thread)
        self._reader.start()

    def encode(self, image: Image) -> bytes:
        return image.convert('RGB').tobytes()

    def _reader_thread(self):
        start_time = time.time()

        while not self.is_closed():
            try:
                data = self.ffmpeg.stdout.read(1 << 15)
            except Exception as e:
                self.logger.warning('FFmpeg reader error: {}'.format(str(e)))
                break

            if not data:
                continue

            if self.frame is None:
                latency = time.time() - start_time
                self.logger.info('FFmpeg stream latency: {} secs'.format(latency))

            with self.ready:
                self.frame = data
                self.frame_time = time.time()
                self.ready.notify_all()

            self._sock_send(self.frame)

    def write(self, image: Image):
        if self.is_closed():
            return

        data = self.encode(image)
        try:
            self.ffmpeg.stdin.write(data)
        except Exception as e:
            self.logger.warning('FFmpeg send error: {}'.format(str(e)))
            self.close()

    def close(self):
        super().close()
        if self._reader and self._reader.is_alive() and threading.get_ident() != self._reader.ident:
            self._reader.join(timeout=5.0)
            self._reader = None


class MKVStreamWriter(FFmpegStreamWriter):
    mimetype = 'video/webm'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, output_format='matroska', **kwargs)


class H264StreamWriter(FFmpegStreamWriter):
    mimetype = 'video/h264'

    def __init__(self, camera: Camera, *args, **kwargs):
        if not camera.info.output_codec:
            camera.info.output_codec = 'libxvid'
        super().__init__(camera, *args, output_format='h264', **kwargs)


class H265StreamWriter(FFmpegStreamWriter):
    mimetype = 'video/h265'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, output_format='h265', **kwargs)


# vim:sw=4:ts=4:et:



import ast
import datetime
import hashlib
import importlib
import inspect
import logging
import os
import pathlib
import re
import rsa
import signal
import socket
import ssl
import urllib.request
from typing import Optional, Tuple, Union

from dateutil import parser, tz
from redis import Redis
from rsa.key import PublicKey, PrivateKey

logger = logging.getLogger('utils')


def get_module_and_method_from_action(action):
    """ Input  : action=music.mpd.play
        Output : ('music.mpd', 'play') """

    tokens = action.split('.')
    module_name = str.join('.', tokens[:-1])
    method_name = tokens[-1:][0]
    return module_name, method_name


def get_message_class_by_type(msgtype):
    """ Gets the class of a message type given as string """

    try:
        module = importlib.import_module('platypush.message.' + msgtype)
    except ImportError as e:
        logger.warning('Unsupported message type {}'.format(msgtype))
        raise RuntimeError(e)

    cls_name = msgtype[0].upper() + msgtype[1:]

    try:
        msgclass = getattr(module, cls_name)
    except AttributeError as e:
        logger.warning('No such class in {}: {}'.format(
            module.__name__, cls_name))
        raise RuntimeError(e)

    return msgclass


# noinspection PyShadowingBuiltins
def get_event_class_by_type(type):
    """ Gets an event class by type name """
    event_module = importlib.import_module('.'.join(type.split('.')[:-1]))
    return getattr(event_module, type.split('.')[-1])


def get_plugin_module_by_name(plugin_name):
    """ Gets the module of a plugin by name (e.g. "music.mpd" or "media.vlc") """

    module_name = 'platypush.plugins.' + plugin_name
    try:
        return importlib.import_module('platypush.plugins.' + plugin_name)
    except ImportError as e:
        logger.error('Cannot import {}: {}'.format(module_name, str(e)))
        return None


def get_plugin_class_by_name(plugin_name):
    """ Gets the class of a plugin by name (e.g. "music.mpd" or "media.vlc") """

    module = get_plugin_module_by_name(plugin_name)
    if not module:
        return

    class_name = getattr(module, ''.join([_.capitalize() for _ in plugin_name.split('.')]) + 'Plugin')
    try:
        return getattr(module, ''.join([_.capitalize() for _ in plugin_name.split('.')]) + 'Plugin')
    except Exception as e:
        logger.error('Cannot import class {}: {}'.format(class_name, str(e)))
        return None


def get_plugin_name_by_class(plugin) -> Optional[str]:
    """Gets the common name of a plugin (e.g. "music.mpd" or "media.vlc") given its class. """

    from platypush.plugins import Plugin

    if isinstance(plugin, Plugin):
        plugin = plugin.__class__

    class_name = plugin.__name__
    class_tokens = [
        token.lower() for token in re.sub(r'([A-Z])', r' \1', class_name).split(' ')
        if token.strip() and token != 'Plugin'
    ]

    return '.'.join(class_tokens)


def get_backend_name_by_class(backend) -> Optional[str]:
    """Gets the common name of a backend (e.g. "http" or "mqtt") given its class. """

    from platypush.backend import Backend

    if isinstance(backend, Backend):
        backend = backend.__class__

    class_name = backend.__name__
    class_tokens = [
        token.lower() for token in re.sub(r'([A-Z])', r' \1', class_name).split(' ')
        if token.strip() and token != 'Backend'
    ]

    return '.'.join(class_tokens)


def set_timeout(seconds, on_timeout):
    """
    Set a function to be called if timeout expires without being cleared.
    It only works on the main thread.

    Params:
        seconds    -- Timeout in seconds
        on_timeout -- Function invoked on timeout unless clear_timeout is called before
    """

    def _sighandler(*_):
        on_timeout()

    signal.signal(signal.SIGALRM, _sighandler)
    signal.alarm(seconds)


def clear_timeout():
    """ Clear any previously set timeout """
    signal.alarm(0)


def get_hash(s):
    """ Get the SHA256 hash hex digest of a string input """
    return hashlib.sha256(s.encode('utf-8')).hexdigest()


def get_decorators(cls, climb_class_hierarchy=False):
    """
    Get the decorators of a class as a {"decorator_name": [list of methods]} dictionary

    :param cls: Class type
    :param climb_class_hierarchy: If set to True (default: False), it will search return the decorators in the parent
        classes as well
    :type climb_class_hierarchy: bool
    """

    decorators = {}

    # noinspection PyPep8Naming
    def visit_FunctionDef(node):
        for n in node.decorator_list:
            if isinstance(n, ast.Call):
                # noinspection PyUnresolvedReferences
                name = n.func.attr if isinstance(n.func, ast.Attribute) else n.func.id
            else:
                name = n.attr if isinstance(n, ast.Attribute) else n.id

            decorators[name] = decorators.get(name, set())
            decorators[name].add(node.name)

    if climb_class_hierarchy:
        targets = inspect.getmro(cls)
    else:
        targets = [cls]

    node_iter = ast.NodeVisitor()
    node_iter.visit_FunctionDef = visit_FunctionDef

    for target in targets:
        try:
            node_iter.visit(ast.parse(inspect.getsource(target)))
        except TypeError:
            # Ignore built-in classes
            pass

    return decorators


def get_redis_queue_name_by_message(msg):
    from platypush.message import Message

    if not isinstance(msg, Message):
        logger.warning('Not a valid message (type: {}): {}'.format(type(msg), msg))

    return 'platypush/responses/{}'.format(msg.id) if msg.id else None


def _get_ssl_context(context_type=None, ssl_cert=None, ssl_key=None,
                     ssl_cafile=None, ssl_capath=None):
    if not context_type:
        ssl_context = ssl.create_default_context(cafile=ssl_cafile,
                                                 capath=ssl_capath)
    else:
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)

    if ssl_cafile or ssl_capath:
        ssl_context.load_verify_locations(
            cafile=ssl_cafile, capath=ssl_capath)

    ssl_context.load_cert_chain(
        certfile=os.path.abspath(os.path.expanduser(ssl_cert)),
        keyfile=os.path.abspath(os.path.expanduser(ssl_key)) if ssl_key else None
    )

    return ssl_context


def get_ssl_context(ssl_cert=None, ssl_key=None, ssl_cafile=None,
                    ssl_capath=None):
    return _get_ssl_context(context_type=None,
                            ssl_cert=ssl_cert, ssl_key=ssl_key,
                            ssl_cafile=ssl_cafile, ssl_capath=ssl_capath)


def get_ssl_server_context(ssl_cert=None, ssl_key=None, ssl_cafile=None,
                           ssl_capath=None):
    return _get_ssl_context(context_type=ssl.PROTOCOL_TLS_SERVER,
                            ssl_cert=ssl_cert, ssl_key=ssl_key,
                            ssl_cafile=ssl_cafile, ssl_capath=ssl_capath)


def get_ssl_client_context(ssl_cert=None, ssl_key=None, ssl_cafile=None,
                           ssl_capath=None):
    return _get_ssl_context(context_type=ssl.PROTOCOL_TLS_CLIENT,
                            ssl_cert=ssl_cert, ssl_key=ssl_key,
                            ssl_cafile=ssl_cafile, ssl_capath=ssl_capath)


def set_thread_name(name):
    global logger

    try:
        import prctl
        # noinspection PyUnresolvedReferences
        prctl.set_name(name)
    except ImportError:
        logger.debug('Unable to set thread name: prctl module is missing')


def find_bins_in_path(bin_name):
    return [
        os.path.join(p, bin_name)
        for p in os.environ.get('PATH', '').split(':')
        if os.path.isfile(os.path.join(p, bin_name)) and (
            os.name == 'nt' or os.access(os.path.join(p, bin_name), os.X_OK)
        )]


def find_files_by_ext(directory, *exts):
    """
    Finds all the files in the given directory with the provided extensions
    """

    if not exts:
        raise AttributeError('No extensions provided')

    if not os.path.isdir(directory):
        raise AttributeError('{} is not a valid directory'.format(directory))

    min_len = len(min(exts, key=len))
    max_len = len(max(exts, key=len))
    result = []

    for root, dirs, files in os.walk(directory):
        for i in range(min_len, max_len + 1):
            result += [f for f in files if f[-i:] in exts]

    return result


def is_process_alive(pid):
    try:
        os.kill(pid, 0)
        return True
    except OSError:
        return False


def get_ip_or_hostname():
    ip = socket.gethostbyname(socket.gethostname())
    if ip.startswith('127.') or ip.startswith('::1'):
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.connect(('10.255.255.255', 1))
            ip = sock.getsockname()[0]
            sock.close()
        except Exception as e:
            logger.debug(e)

    return ip


def get_mime_type(resource):
    import magic
    if resource.startswith('file://'):
        resource = resource[len('file://'):]

    # noinspection HttpUrlsUsage
    if resource.startswith('http://') or resource.startswith('https://'):
        with urllib.request.urlopen(resource) as response:
            return response.info().get_content_type()
    else:
        if hasattr(magic, 'detect_from_filename'):
            mime = magic.detect_from_filename(resource)
        elif hasattr(magic, 'from_file'):
            mime = magic.from_file(resource, mime=True)
        else:
            raise RuntimeError('The installed magic version provides neither detect_from_filename nor from_file')

        if mime:
            return mime.mime_type if hasattr(mime, 'mime_type') else mime


def camel_case_to_snake_case(string):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', string)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def grouper(n, iterable, fillvalue=None):
    """
    Split an iterable in groups of max N elements.
    grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx
    """
    from itertools import zip_longest
    args = [iter(iterable)] * n

    if fillvalue:
        return zip_longest(fillvalue=fillvalue, *args)

    for chunk in zip_longest(*args):
        yield filter(None, chunk)


def is_functional_procedure(obj) -> bool:
    return callable(obj) and hasattr(obj, 'procedure')


def is_functional_hook(obj) -> bool:
    return callable(obj) and hasattr(obj, 'hook')


def is_functional_cron(obj) -> bool:
    return callable(obj) and hasattr(obj, 'cron') and hasattr(obj, 'cron_expression')


def run(action, *args, **kwargs):
    from platypush.context import get_plugin
    (module_name, method_name) = get_module_and_method_from_action(action)
    plugin = get_plugin(module_name)
    method = getattr(plugin, method_name)
    response = method(*args, **kwargs)

    if response.errors:
        raise RuntimeError(response.errors[0])

    return response.output


def generate_rsa_key_pair(key_file: Optional[str] = None, size: int = 2048) \
        -> Tuple[PublicKey, PrivateKey]:
    """
    Generate an RSA key pair.

    :param key_file: Target file for the private key (the associated public key will be stored in ``<key_file>.pub``.
        If no key file is specified then the public and private keys will be returned in ASCII format in a dictionary
        with the following structure:

            .. code-block:: json

                {
                    "private": "private key here",
                    "public": "public key here"
                }

    :param size: Key size (default: 2048 bits).
    :return: A tuple with the generated ``(priv_key_str, pub_key_str)``.
    """
    logger.info('Generating RSA keypair')
    pub_key, priv_key = rsa.newkeys(size)
    logger.info('Generated RSA keypair')
    public_key_str = pub_key.save_pkcs1('PEM').decode()
    private_key_str = priv_key.save_pkcs1('PEM').decode()

    if key_file:
        logger.info('Saving private key to {}'.format(key_file))
        with open(os.path.expanduser(key_file), 'w') as f1, \
                open(os.path.expanduser(key_file) + '.pub', 'w') as f2:
            f1.write(private_key_str)
            f2.write(public_key_str)
            os.chmod(key_file, 0o600)

    return pub_key, priv_key


def get_or_generate_jwt_rsa_key_pair():
    from platypush.config import Config

    key_dir = os.path.join(Config.get('workdir'), 'jwt')
    priv_key_file = os.path.join(key_dir, 'id_rsa')
    pub_key_file = priv_key_file + '.pub'

    if os.path.isfile(priv_key_file) and os.path.isfile(pub_key_file):
        with (
            open(pub_key_file, 'r') as f1,
            open(priv_key_file, 'r') as f2
        ):
            return (
                rsa.PublicKey.load_pkcs1(f1.read().encode()),
                rsa.PrivateKey.load_pkcs1(f2.read().encode()),
            )

    pathlib.Path(key_dir).mkdir(parents=True, exist_ok=True, mode=0o755)
    return generate_rsa_key_pair(priv_key_file, size=2048)


def get_enabled_plugins() -> dict:
    from platypush.config import Config
    from platypush.context import get_plugin

    plugins = {}
    for name, config in Config.get_plugins().items():
        try:
            plugin = get_plugin(name)
            if plugin:
                plugins[name] = plugin
        except Exception as e:
            logger.warning(f'Could not initialize plugin {name}')
            logger.exception(e)

    return plugins


def get_redis() -> Redis:
    from platypush.config import Config
    return Redis(**(Config.get('backend.redis') or {}).get('redis_args', {}))


def to_datetime(t: Union[str, int, float, datetime.datetime]) -> datetime.datetime:
    if isinstance(t, int) or isinstance(t, float):
        return datetime.datetime.fromtimestamp(t, tz=tz.tzutc())
    if isinstance(t, str):
        return parser.parse(t)
    return t


# vim:sw=4:ts=4:et:



import logging
import subprocess
import threading

from platypush.plugins.camera.model.writer.image import MJPEGStreamWriter
from platypush.plugins.camera.model.writer.preview import PreviewWriter


class FFplayPreviewWriter(PreviewWriter, MJPEGStreamWriter):
    """
    General class for managing previews from camera devices or generic sources of images over ffplay.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(self.__class__.__name__)
        self.ffplay = subprocess.Popen(['ffplay', '-'], stdin=subprocess.PIPE)
        self._preview_thread = threading.Thread(target=self._ffplay_thread)
        self._preview_thread.start()

    def _ffplay_thread(self):
        while not self.closed and self.ffplay.poll() is None:
            with self.ready:
                self.ready.wait(1.)
                if not self.frame:
                    continue

                try:
                    self.ffplay.stdin.write(self.frame)
                except Exception as e:
                    self.logger.warning('ffplay write error: {}'.format(str(e)))
                    self.close()
                    break

    def close(self):
        if self.ffplay and self.ffplay.poll() is None:
            self.ffplay.terminate()

        self.camera = None
        super().close()
        if self._preview_thread and self._preview_thread.is_alive() and \
                threading.get_ident() != self._preview_thread.ident:
            self._preview_thread.join(timeout=5.0)
            self._preview_thread = None


# vim:sw=4:ts=4:et:



import datetime
import os
from typing import Optional, Union, List, Dict, Any

from sqlalchemy import create_engine, Column, Integer, String, DateTime
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.ext.declarative import declarative_base

from platypush.backend import Backend
from platypush.config import Config
from platypush.context import get_plugin
from platypush.message.event.covid19 import Covid19UpdateEvent
from platypush.plugins.covid19 import Covid19Plugin

Base = declarative_base()
Session = scoped_session(sessionmaker())


class Covid19Update(Base):
    """ Models the Covid19Data table """

    __tablename__ = 'covid19data'
    __table_args__ = ({'sqlite_autoincrement': True})

    country = Column(String, primary_key=True)
    confirmed = Column(Integer, nullable=False, default=0)
    deaths = Column(Integer, nullable=False, default=0)
    recovered = Column(Integer, nullable=False, default=0)
    last_updated_at = Column(DateTime, nullable=False, default=datetime.datetime.utcnow)


class Covid19Backend(Backend):
    """
    This backend polls new data about the Covid-19 pandemic diffusion and triggers events when new data is available.

    Triggers:

        - :class:`platypush.message.event.covid19.Covid19UpdateEvent` when new data is available.

    """

    # noinspection PyProtectedMember
    def __init__(self, country: Optional[Union[str, List[str]]], poll_seconds: Optional[float] = 3600.0, **kwargs):
        """
        :param country: Default country (or list of countries) to retrieve the stats for. It can either be the full
            country name or the country code. Special values:

            - ``world``: Get worldwide stats.
            - ``all``: Get all the available stats.

        Default: either the default configured on the :class:`platypush.plugins.covid19.Covid19Plugin` plugin or
        ``world``.

        :param poll_seconds: How often the backend should check for new check-ins (default: one hour).
        """
        super().__init__(poll_seconds=poll_seconds, **kwargs)
        self._plugin: Covid19Plugin = get_plugin('covid19')
        self.country: List[str] = self._plugin._get_countries(country)
        self.workdir = os.path.join(os.path.expanduser(Config.get('workdir')), 'covid19')
        self.dbfile = os.path.join(self.workdir, 'data.db')
        os.makedirs(self.workdir, exist_ok=True)

    def __enter__(self):
        self.logger.info('Started Covid19 backend')

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logger.info('Stopped Covid19 backend')

    def _process_update(self, summary: Dict[str, Any], session: Session):
        update_time = datetime.datetime.fromisoformat(summary['Date'].replace('Z', '+00:00'))

        self.bus.post(Covid19UpdateEvent(
            country=summary['Country'],
            country_code=summary['CountryCode'],
            confirmed=summary['TotalConfirmed'],
            deaths=summary['TotalDeaths'],
            recovered=summary['TotalRecovered'],
            update_time=update_time,
        ))

        session.merge(Covid19Update(country=summary['CountryCode'],
                                    confirmed=summary['TotalConfirmed'],
                                    deaths=summary['TotalDeaths'],
                                    recovered=summary['TotalRecovered'],
                                    last_updated_at=update_time))

    def loop(self):
        # noinspection PyUnresolvedReferences
        summaries = self._plugin.summary(self.country).output
        if not summaries:
            return

        engine = create_engine('sqlite:///{}'.format(self.dbfile), connect_args={'check_same_thread': False})
        Base.metadata.create_all(engine)
        Session.configure(bind=engine)
        session = Session()

        last_records = {
            record.country: record
            for record in session.query(Covid19Update).filter(Covid19Update.country.in_(self.country)).all()
        }

        for summary in summaries:
            country = summary['CountryCode']
            last_record = last_records.get(country)
            if not last_record or \
                    summary['TotalConfirmed'] != last_record.confirmed or \
                    summary['TotalDeaths'] != last_record.deaths or \
                    summary['TotalRecovered'] != last_record.recovered:
                self._process_update(summary=summary, session=session)

        session.commit()


# vim:sw=4:ts=4:et:



import inspect
import logging
import os

from platypush.utils.manifest import Manifest

logger = logging.getLogger('platypush')


def exec_wrapper(f, *args, **kwargs):
    from platypush import Response

    try:
        ret = f(*args, **kwargs)
        if isinstance(ret, Response):
            return ret

        return Response(output=ret)
    except Exception as e:
        logger.exception(e)
        return Response(errors=[str(e)])


class ExtensionWithManifest:
    def __init__(self, *_, **__):
        self._manifest = self.get_manifest()

    def get_manifest(self) -> Manifest:
        manifest_file = os.path.join(os.path.dirname(inspect.getfile(self.__class__)), 'manifest.yaml')
        assert os.path.isfile(manifest_file), (
            'The extension {} has no associated manifest.yaml'.format(self.__class__.__name__)
        )

        return Manifest.from_file(manifest_file)



import json
import os
import pathlib
import requests
import time

from datetime import datetime, timedelta
from typing import Iterable, List, Optional
from urllib.parse import urljoin

from platypush.config import Config
from platypush.plugins import Plugin, action
from platypush.schemas.wallabag import WallabagEntrySchema


class WallabagPlugin(Plugin):
    """
    Plugin to interact with Wallabag (https://wallabag.it),
    an open-source alternative to Instapaper and Pocket.
    """

    _default_credentials_file = os.path.join(
        str(Config.get('workdir')), 'wallabag', 'credentials.json'
    )

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        server_url: str = 'https://wallabag.it',
        username: Optional[str] = None,
        password: Optional[str] = None,
        credentials_file: str = _default_credentials_file,
        **kwargs,
    ):
        """
        :param client_id: Client ID for your application - you can create one
            at ``<server_url>/developer``.
        :param client_secret: Client secret for your application - you can
            create one at ``<server_url>/developer``.
        :param server_url: Base URL of the Wallabag server (default: ``https://wallabag.it``).
        :param username: Wallabag username. Only needed for the first login,
            you can remove it afterwards. Alternatively, you can provide it
            on the :meth:`.login` method.
        :param password: Wallabag password. Only needed for the first login,
            you can remove it afterwards. Alternatively, you can provide it
            on the :meth:`.login` method.
        :param credentials_file: Path to the file where the OAuth session
            parameters will be stored (default:
            ``<WORKDIR>/wallabag/credentials.json``).
        """
        super().__init__(**kwargs)
        self._client_id = client_id
        self._client_secret = client_secret
        self._server_url = server_url
        self._username = username
        self._password = password
        self._credentials_file = os.path.expanduser(credentials_file)
        self._session = {}

    def _oauth_open_saved_session(self):
        try:
            with open(self._credentials_file, 'r') as f:
                data = json.load(f)
        except Exception as e:
            self.logger.warning('Could not load %s: %s', self._credentials_file, e)
            return

        self._session = {
            'username': data['username'],
            'client_id': data.get('client_id', self._client_id),
            'client_secret': data.get('client_secret', self._client_secret),
            'access_token': data['access_token'],
            'refresh_token': data['refresh_token'],
        }

        if data.get('expires_at') and time.time() > data['expires_at']:
            self.logger.info('OAuth token expired, refreshing it')
            self._oauth_refresh_token()

    def _oauth_refresh_token(self):
        url = urljoin(self._server_url, '/oauth/v2/token')
        rs = requests.post(
            url,
            json={
                'grant_type': 'refresh_token',
                'client_id': self._client_id,
                'client_secret': self._client_secret,
                'access_token': self._session['access_token'],
                'refresh_token': self._session['refresh_token'],
            },
        )

        rs.raise_for_status()
        rs = rs.json()
        self._session.update(
            {
                'access_token': rs['access_token'],
                'refresh_token': rs['refresh_token'],
                'expires_at': (
                    int(
                        (
                            datetime.now() + timedelta(seconds=rs['expires_in'])
                        ).timestamp()
                    )
                    if rs.get('expires_in')
                    else None
                ),
            }
        )

        self._oauth_flush_session()

    def _oauth_create_new_session(self, username: str, password: str):
        url = urljoin(self._server_url, '/oauth/v2/token')
        rs = requests.post(
            url,
            json={
                'grant_type': 'password',
                'client_id': self._client_id,
                'client_secret': self._client_secret,
                'username': username,
                'password': password,
            },
        )

        rs.raise_for_status()
        rs = rs.json()
        self._session = {
            'client_id': self._client_id,
            'client_secret': self._client_secret,
            'username': username,
            'access_token': rs['access_token'],
            'refresh_token': rs['refresh_token'],
            'expires_at': (
                int((datetime.now() + timedelta(seconds=rs['expires_in'])).timestamp())
                if rs.get('expires_in')
                else None
            ),
        }

        self._oauth_flush_session()

    def _oauth_flush_session(self):
        pathlib.Path(self._credentials_file).parent.mkdir(parents=True, exist_ok=True)

        pathlib.Path(self._credentials_file).touch(mode=0o600, exist_ok=True)
        with open(self._credentials_file, 'w') as f:
            f.write(json.dumps(self._session))

    @action
    def login(self, username: Optional[str] = None, password: Optional[str] = None):
        """
        Create a new user session if not logged in.

        :param username: Default ``username`` override.
        :param password: Default ``password`` override.
        """
        self._oauth_open_saved_session()
        if self._session:
            return

        username = username or self._username
        password = password or self._password
        assert (
            username and password
        ), 'No stored user session and no username/password provided'

        self._oauth_create_new_session(username, password)

    def _request(self, url: str, method: str, *args, as_json=True, **kwargs):
        url = urljoin(self._server_url, f'api/{url}')
        func = getattr(requests, method.lower())
        self.login()
        kwargs['headers'] = {
            **kwargs.get('headers', {}),
            'Authorization': f'Bearer {self._session["access_token"]}',
        }

        rs = func(url, *args, **kwargs)
        rs.raise_for_status()
        return rs.json() if as_json else rs.text

    @action
    def list(
        self,
        archived: bool = True,
        starred: bool = False,
        sort: str = 'created',
        descending: bool = False,
        page: int = 1,
        limit: int = 30,
        tags: Optional[Iterable[str]] = None,
        since: Optional[int] = None,
        full: bool = True,
    ) -> List[dict]:
        """
        List saved links.

        :param archived: Include archived items (default: ``True``).
        :param starred: Include only starred items (default: ``False``).
        :param sort: Timestamp sort criteria. Supported: ``created``,
            ``updated``, ``archived`` (default: ``created``).
        :param descending: Sort in descending order (default: ``False``).
        :param page: Results page to be retrieved (default: ``1``).
        :param limit: Maximum number of entries per page (default: ``30``).
        :param tags: Filter by a list of tags.
        :param since: Return entries created after this timestamp (as a UNIX
            timestamp).
        :param full: Include the full parsed body of the saved entry.
        :return: .. schema:: wallabag.WallabagEntrySchema(many=True)
        """
        rs = self._request(
            '/entries.json',
            method='get',
            params={
                'archived': int(archived),
                'starred': int(starred),
                'sort': sort,
                'order': 'desc' if descending else 'asc',
                'page': page,
                'perPage': limit,
                'tags': ','.join(tags or []),
                'since': since or 0,
                'detail': 'full' if full else 'metadata',
            },
        )

        return WallabagEntrySchema().dump(
            rs.get('_embedded', {}).get('items', []), many=True
        )

    @action
    def search(
        self,
        term: str,
        page: int = 1,
        limit: int = 30,
    ) -> List[dict]:
        """
        Search links by some text.

        :param term: Term to be searched.
        :param page: Results page to be retrieved (default: ``1``).
        :param limit: Maximum number of entries per page (default: ``30``).
        :return: .. schema:: wallabag.WallabagEntrySchema(many=True)
        """
        rs = self._request(
            '/search.json',
            method='get',
            params={
                'term': term,
                'page': page,
                'perPage': limit,
            },
        )

        return WallabagEntrySchema().dump(
            rs.get('_embedded', {}).get('items', []), many=True
        )

    @action
    def get(self, id: int) -> Optional[dict]:
        """
        Get the content and metadata of a link by ID.

        :param id: Entry ID.
        :return: .. schema:: wallabag.WallabagEntrySchema
        """
        rs = self._request(f'/entries/{id}.json', method='get')
        return WallabagEntrySchema().dump(rs)  # type: ignore

    @action
    def export(self, id: int, file: str, format: str = 'txt'):
        """
        Export a saved entry to a file in the specified format.

        :param id: Entry ID.
        :param file: Output filename.
        :param format: Output format. Supported: ``txt``, ``xml``, ``csv``,
            ``pdf``, ``epub`` and ``mobi`` (default: ``txt``).
        """
        rs = self._request(
            f'/entries/{id}/export.{format}', method='get', as_json=False
        )

        if isinstance(rs, str):
            rs = rs.encode()
        with open(os.path.expanduser(file), 'wb') as f:
            f.write(rs)

    @action
    def save(
        self,
        url: str,
        title: Optional[str] = None,
        content: Optional[str] = None,
        tags: Optional[Iterable[str]] = None,
        authors: Optional[Iterable[str]] = None,
        archived: bool = False,
        starred: bool = False,
        public: bool = False,
        language: Optional[str] = None,
        preview_picture: Optional[str] = None,
    ) -> Optional[dict]:
        """
        Save a link to Wallabag.

        :param url: URL to be saved.
        :param title: Entry title (default: parsed from the page content).
        :param content: Entry content (default: parsed from the entry itself).
        :param tags: List of tags to attach to the entry.
        :param authors: List of authors of the entry (default: parsed from the
            page content).
        :param archived: Whether the entry should be created in the archive
            (default: ``False``).
        :param starred: Whether the entry should be starred (default:
            ``False``).
        :param public: Whether the entry should be publicly available. If so, a
            public URL will be generated (default: ``False``).
        :param language: Language of the entry.
        :param preview_picture: URL of a picture to be used for the preview
            (default: parsed from the page itself).
        :return: .. schema:: wallabag.WallabagEntrySchema
        """
        rs = self._request(
            '/entries.json',
            method='post',
            json={
                'url': url,
                'title': title,
                'content': content,
                'tags': ','.join(tags or []),
                'authors': ','.join(authors or []),
                'archive': int(archived),
                'starred': int(starred),
                'public': int(public),
                'language': language,
                'preview_picture': preview_picture,
            },
        )

        return WallabagEntrySchema().dump(rs)  # type: ignore

    @action
    def update(
        self,
        id: int,
        title: Optional[str] = None,
        content: Optional[str] = None,
        tags: Optional[Iterable[str]] = None,
        authors: Optional[Iterable[str]] = None,
        archived: bool = False,
        starred: bool = False,
        public: bool = False,
        language: Optional[str] = None,
        preview_picture: Optional[str] = None,
    ) -> Optional[dict]:
        """
        Update a link entry saved to Wallabag.

        :param id: Entry ID.
        :param title: New entry title.
        :param content: New entry content.
        :param tags: List of tags to attach to the entry.
        :param authors: List of authors of the entry.
        :param archived: Archive/unarchive the entry.
        :param starred: Star/unstar the entry.
        :param public: Mark the entry as public/private.
        :param language: Change the language of the entry.
        :param preview_picture: Change the preview picture URL.
        :return: .. schema:: wallabag.WallabagEntrySchema
        """
        rs = self._request(
            f'/entries/{id}.json',
            method='patch',
            json={
                'title': title,
                'content': content,
                'tags': ','.join(tags or []),
                'authors': ','.join(authors or []),
                'archive': int(archived),
                'starred': int(starred),
                'public': int(public),
                'language': language,
                'preview_picture': preview_picture,
            },
        )

        return WallabagEntrySchema().dump(rs)  # type: ignore

    @action
    def delete(self, id: int) -> Optional[dict]:
        """
        Delete an entry by ID.

        :param id: Entry ID.
        :return: .. schema:: wallabag.WallabagEntrySchema
        """
        rs = self._request(
            f'/entries/{id}.json',
            method='delete',
        )

        return WallabagEntrySchema().dump(rs)  # type: ignore



from dataclasses import dataclass
from typing import Optional, List


@dataclass
class MonitoredResource:
    path: str
    recursive: bool = False


@dataclass
class MonitoredPattern(MonitoredResource):
    patterns: Optional[List[str]] = None
    ignore_patterns: Optional[List[str]] = None
    ignore_directories: Optional[List[str]] = None
    case_sensitive: bool = True


@dataclass
class MonitoredRegex(MonitoredResource):
    regexes: Optional[List[str]] = None
    ignore_regexes: Optional[List[str]] = None
    ignore_directories: Optional[List[str]] = None
    case_sensitive: bool = True



import numpy as np
from PIL.Image import Image as ImageType

from platypush.plugins.camera.model.writer import FileVideoWriter


class CvFileWriter(FileVideoWriter):
    """
    Write camera frames to a file using OpenCV.
    """
    def __init__(self, *args, **kwargs):
        import cv2
        super(CvFileWriter, self).__init__(*args, **kwargs)

        video_type = cv2.VideoWriter_fourcc(*(self.camera.info.output_format or 'xvid').upper())
        resolution = (
            int(self.camera.info.resolution[0] * (self.camera.info.scale_x or 1.)),
            int(self.camera.info.resolution[1] * (self.camera.info.scale_y or 1.)),
        )

        self.writer = cv2.VideoWriter(self.output_file, video_type, self.camera.info.fps, resolution, False)

    def write(self, img):
        if not self.writer:
            return

        try:
            if isinstance(img, ImageType):
                # noinspection PyTypeChecker
                img = np.array(img)
        except (ValueError, TypeError):
            pass

        self.writer.write(img)

    def close(self):
        if not self.writer:
            return

        self.writer.release()
        self.writer = None
        super().close()


# vim:sw=4:ts=4:et:



import logging

from abc import ABC

from platypush.plugins.camera.model.writer import VideoWriter

logger = logging.getLogger('platypush')


class PreviewWriter(VideoWriter, ABC):
    """
    Abstract class for camera previews.
    """


class PreviewWriterFactory:
    @staticmethod
    def get(*args, **kwargs) -> PreviewWriter:
        try:
            import wx
            # noinspection PyUnresolvedReferences
            from platypush.plugins.camera.model.writer.preview.wx import WxPreviewWriter
            return WxPreviewWriter(*args, **kwargs)
        except ImportError:
            logger.warning('wxPython not available, using ffplay as a fallback for camera previews')

        from platypush.plugins.camera.model.writer.preview.ffplay import FFplayPreviewWriter
        return FFplayPreviewWriter(*args, **kwargs)


# vim:sw=4:ts=4:et:



from flask import Blueprint

from platypush.backend.http.app import template_folder
from platypush.backend.http.app.routes.plugins.camera import get_photo, get_video
from platypush.backend.http.app.utils import authenticate

camera_ir_mlx90640 = Blueprint('camera-ir-mlx90640', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    camera_ir_mlx90640,
]


@camera_ir_mlx90640.route('/camera/ir/mlx90640/photo.<extension>', methods=['GET'])
@authenticate()
def get_photo_route(extension):
    return get_photo('ir.mlx90640', extension)


@camera_ir_mlx90640.route('/camera/ir/mlx90640/video.<extension>', methods=['GET'])
@authenticate()
def get_video_route(extension):
    return get_video('ir.mlx90640', extension)


@camera_ir_mlx90640.route('/camera/ir/mlx90640/photo', methods=['GET'])
@authenticate()
def get_photo_route_default():
    return get_photo_route('jpeg')


@camera_ir_mlx90640.route('/camera/ir/mlx90640/video', methods=['GET'])
@authenticate()
def get_video_route_default():
    return get_video_route('mjpeg')


@camera_ir_mlx90640.route('/camera/ir/mlx90640/frame', methods=['GET'])
@authenticate()
def get_photo_route_deprecated():
    return get_photo_route_default()


@camera_ir_mlx90640.route('/camera/ir/mlx90640/feed', methods=['GET'])
@authenticate()
def get_video_route_deprecated():
    return get_video_route_default()


# vim:sw=4:ts=4:et:



import asyncio
import aiohttp

from threading import RLock
from typing import Optional, Dict, List, Union

from platypush.plugins import action
from platypush.plugins.switch import SwitchPlugin


class SmartthingsPlugin(SwitchPlugin):
    """
    Plugin to interact with devices and locations registered to a Samsung SmartThings account.

    Requires:

        * **pysmartthings** (``pip install pysmartthings``)

    """

    _timeout = aiohttp.ClientTimeout(total=20.)

    def __init__(self, access_token: str, **kwargs):
        """
        :param access_token: SmartThings API access token - you can get one at https://account.smartthings.com/tokens.
        """
        super().__init__(**kwargs)
        self._access_token = access_token
        self._refresh_lock = RLock()
        self._execute_lock = RLock()

        self._locations = []
        self._devices = []
        self._rooms_by_location = {}

        self._locations_by_id = {}
        self._locations_by_name = {}
        self._devices_by_id = {}
        self._devices_by_name = {}
        self._rooms_by_id = {}
        self._rooms_by_location_and_id = {}
        self._rooms_by_location_and_name = {}

    async def _refresh_locations(self, api):
        self._locations = await api.locations()

        self._locations_by_id = {
            loc.location_id: loc
            for loc in self._locations
        }

        self._locations_by_name = {
            loc.name: loc
            for loc in self._locations
        }

    async def _refresh_devices(self, api):
        self._devices = await api.devices()

        self._devices_by_id = {
            dev.device_id: dev
            for dev in self._devices
        }

        self._devices_by_name = {
            dev.label: dev
            for dev in self._devices
        }

    async def _refresh_rooms(self, api, location_id: str):
        self._rooms_by_location[location_id] = await api.rooms(location_id=location_id)

        self._rooms_by_id.update(**{
            room.room_id: room
            for room in self._rooms_by_location[location_id]
        })

        self._rooms_by_location_and_id[location_id] = {
            room.room_id: room
            for room in self._rooms_by_location[location_id]
        }

        self._rooms_by_location_and_name[location_id] = {
            room.name: room
            for room in self._rooms_by_location[location_id]
        }

    async def _refresh_info(self):
        import pysmartthings

        async with aiohttp.ClientSession(timeout=self._timeout) as session:
            api = pysmartthings.SmartThings(session, self._access_token)
            tasks = [
                asyncio.ensure_future(self._refresh_locations(api)),
                asyncio.ensure_future(self._refresh_devices(api)),
            ]

            await asyncio.gather(*tasks)

            room_tasks = [
                asyncio.ensure_future(self._refresh_rooms(api, location.location_id))
                for location in self._locations
            ]

            await asyncio.gather(*room_tasks)

    def refresh_info(self):
        with self._refresh_lock:
            loop = asyncio.new_event_loop()
            try:
                asyncio.set_event_loop(loop)
                loop.run_until_complete(self._refresh_info())
            finally:
                loop.stop()

    def _location_to_dict(self, location) -> Dict:
        return {
            'name': location.name,
            'location_id': location.location_id,
            'country_code': location.country_code,
            'locale': location.locale,
            'latitude': location.latitude,
            'longitude': location.longitude,
            'temperature_scale': location.temperature_scale,
            'region_radius': location.region_radius,
            'timezone_id': location.timezone_id,
            'rooms': {
                room.room_id: self._room_to_dict(room)
                for room in self._rooms_by_location.get(location.location_id, {})
            }
        }

    @staticmethod
    def _device_to_dict(device) -> Dict:
        return {
            'capabilities': device.capabilities,
            'name': device.label,
            'device_id': device.device_id,
            'location_id': device.location_id,
            'room_id': device.room_id,
            'device_type_id': device.device_type_id,
            'device_type_name': device.device_type_name,
            'device_type_network': device.device_type_network,
        }

    @staticmethod
    def _room_to_dict(room) -> Dict:
        return {
            'name': room.name,
            'background_image': room.background_image,
            'room_id': room.room_id,
            'location_id': room.location_id,
        }

    @action
    def info(self) -> Dict[str, Dict[str, dict]]:
        """
        Return the objects registered to the account, including locations and devices.

          .. code-block:: json

            {
                "devices": {
                    "smart-tv-id": {
                        "capabilities": [
                            "ocf",
                            "switch",
                            "audioVolume",
                            "audioMute",
                            "tvChannel",
                            "mediaInputSource",
                            "mediaPlayback",
                            "mediaTrackControl",
                            "custom.error",
                            "custom.picturemode",
                            "custom.soundmode",
                            "custom.accessibility",
                            "custom.launchapp",
                            "custom.recording",
                            "custom.tvsearch",
                            "custom.disabledCapabilities",
                            "samsungvd.ambient",
                            "samsungvd.ambientContent",
                            "samsungvd.ambient18",
                            "samsungvd.mediaInputSource",
                            "refresh",
                            "execute",
                            "samsungvd.firmwareVersion",
                            "samsungvd.supportsPowerOnByOcf"
                        ],
                        "device_id": "smart-tv-id",
                        "device_type_id": null,
                        "device_type_name": null,
                        "device_type_network": null,
                        "location_id": "location-id",
                        "name": "Samsung Smart TV",
                        "room_id": "room-1"
                    },
                    "tv-switch-id": {
                        "capabilities": [
                            "switch",
                            "refresh",
                            "healthCheck"
                        ],
                        "device_id": "tv-switch-id",
                        "device_type_id": null,
                        "device_type_name": null,
                        "device_type_network": null,
                        "location_id": "location-id",
                        "name": "TV Smart Switch",
                        "room_id": "room-1"
                    },
                    "lights-switch-id": {
                        "capabilities": [
                            "switch",
                            "refresh",
                            "healthCheck"
                        ],
                        "device_id": "lights-switch-id",
                        "device_type_id": null,
                        "device_type_name": null,
                        "device_type_network": null,
                        "location_id": "location-id",
                        "name": "Lights Switch",
                        "room_id": "room-2"
                    }
                },
                "locations": {
                    "location-id": {
                        "name": "My home",
                        "location_id": "location-id",
                        "country_code": "us",
                        "locale": "en-US",
                        "latitude": "latitude",
                        "longitude": "longitude",
                        "temperature_scale": null,
                        "region_radius": null,
                        "timezone_id": null,
                        "rooms": {
                            "room-1": {
                                "background_image": null,
                                "location_id": "location-1",
                                "name": "Living Room",
                                "room_id": "room-1"
                            },
                            "room-2": {
                                "background_image": null,
                                "location_id": "location-1",
                                "name": "Bedroom",
                                "room_id": "room-2"
                            }
                        }
                    }
                }
            }

        """
        self.refresh_info()
        return {
            'locations': {loc.location_id: self._location_to_dict(loc) for loc in self._locations},
            'devices': {dev.device_id: self._device_to_dict(dev) for dev in self._devices},
        }

    @action
    def get_location(self, location_id: Optional[str] = None, name: Optional[str] = None) -> dict:
        """
        Get the info of a location by ID or name.

          .. code-block:: json

            {
                "name": "My home",
                "location_id": "location-id",
                "country_code": "us",
                "locale": "en-US",
                "latitude": "latitude",
                "longitude": "longitude",
                "temperature_scale": null,
                "region_radius": null,
                "timezone_id": null,
                "rooms": {
                    "room-1": {
                        "background_image": null,
                        "location_id": "location-1",
                        "name": "Living Room",
                        "room_id": "room-1"
                    },
                    "room-2": {
                        "background_image": null,
                        "location_id": "location-1",
                        "name": "Bedroom",
                        "room_id": "room-2"
                    }
                }
            }

        """
        assert location_id or name, 'Specify either location_id or name'
        if location_id not in self._locations_by_id or name not in self._locations_by_name:
            self.refresh_info()

        location = self._locations_by_id.get(location_id, self._locations_by_name.get(name))
        assert location, 'Location {} not found'.format(location_id or name)
        return self._location_to_dict(location)

    def _get_device(self, device: str):
        if device not in self._devices_by_id or device not in self._devices_by_name:
            self.refresh_info()

        device = self._devices_by_id.get(device, self._devices_by_name.get(device))
        assert device, 'Device {} not found'.format(device)
        return device

    @action
    def get_device(self, device: str) -> dict:
        """
        Get a device info by ID or name.

        :param device: Device ID or name.
        :return:

          .. code-block:: json

            "tv-switch-id": {
                "capabilities": [
                    "switch",
                    "refresh",
                    "healthCheck"
                ],
                "device_id": "tv-switch-id",
                "device_type_id": null,
                "device_type_name": null,
                "device_type_network": null,
                "location_id": "location-id",
                "name": "TV Smart Switch",
                "room_id": "room-1"
            }

        """
        device = self._get_device(device)
        return self._device_to_dict(device)

    async def _execute(self, device_id: str, capability: str, command, component_id: str, args: Optional[list]):
        import pysmartthings

        async with aiohttp.ClientSession(timeout=self._timeout) as session:
            api = pysmartthings.SmartThings(session, self._access_token)
            device = await api.device(device_id)
            ret = await device.command(component_id=component_id, capability=capability, command=command, args=args)

        assert ret, 'The command {capability}={command} failed on device {device}'.format(
            capability=capability, command=command, device=device_id)

    @action
    def execute(self,
                device: str,
                capability: str,
                command,
                component_id: str = 'main',
                args: Optional[list] = None):
        """
        Execute a command on a device.

        Example request to turn on a device with ``switch`` capability:

          .. code-block:: json

            {
              "type": "request",
              "action": "smartthings.execute",
              "args": {
                "device": "My Switch",
                "capability": "switch",
                "command": "on"
              }
            }

        :param device: Device ID or name.
        :param capability: Property to be read/written (see device ``capabilities`` returned from :meth:`.get_device`).
        :param command: Command to execute on the ``capability``
            (see https://smartthings.developer.samsung.com/docs/api-ref/capabilities.html).
        :param component_id: ID of the component to execute the command on (default: ``main``, i.e. the device itself).
        :param args: Command extra arguments, as a list.
        """
        device = self._get_device(device)

        with self._execute_lock:
            loop = asyncio.new_event_loop()
            try:
                asyncio.set_event_loop(loop)
                loop.run_until_complete(self._execute(
                    device_id=device.device_id, capability=capability, command=command,
                    component_id=component_id, args=args))
            finally:
                loop.stop()

    @staticmethod
    async def _get_device_status(api, device_id: str) -> dict:
        device = await api.device(device_id)
        await device.status.refresh()

        return {
            'device_id': device_id,
            'name': device.label,
            **{
                cap: getattr(device.status, cap)
                for cap in device.capabilities
                if hasattr(device.status, cap)
                and not callable(getattr(device.status, cap))
            }
        }

    async def _refresh_status(self, devices: List[str]) -> List[dict]:
        import pysmartthings

        device_ids = []
        missing_device_ids = set()

        def parse_device_id(device):
            device_id = None
            if device in self._devices_by_id:
                device_id = device
                device_ids.append(device_id)
            elif device in self._devices_by_name:
                device_id = self._devices_by_name[device].device_id
                device_ids.append(device_id)
            else:
                missing_device_ids.add(device)

            if device_id and device in missing_device_ids:
                missing_device_ids.remove(device)

        for dev in devices:
            parse_device_id(dev)

        # Fail if some devices haven't been found after refreshing
        assert not missing_device_ids, 'Could not find the following devices: {}'.format(list(missing_device_ids))

        async with aiohttp.ClientSession(timeout=self._timeout) as session:
            api = pysmartthings.SmartThings(session, self._access_token)
            status_tasks = [
                asyncio.ensure_future(self._get_device_status(api, device_id))
                for device_id in device_ids
            ]

            # noinspection PyTypeChecker
            return await asyncio.gather(*status_tasks)

    @action
    def status(self, device: Optional[Union[str, List[str]]] = None) -> List[dict]:
        """
        Refresh and return the status of one or more devices.

        :param device: Device or list of devices to refresh (default: all)
        :return: A list containing on entry per device, and each entry containing the current device state. Example:

          .. code-block:: json

            [
              {
                "device_id": "switch-1",
                "name": "Fan",
                "switch": false
              },
              {
                "device_id": "tv-1",
                "name": "Samsung Smart TV",
                "switch": true
              }
            ]

        """
        self.refresh_info()

        if not device:
            self.refresh_info()
            devices = self._devices_by_id.keys()
        elif isinstance(device, str):
            devices = [device]
        else:
            devices = device

        with self._refresh_lock:
            loop = asyncio.new_event_loop()
            try:
                asyncio.set_event_loop(loop)
                return loop.run_until_complete(self._refresh_status(devices))
            finally:
                loop.stop()

    @action
    def on(self, device: str, *args, **kwargs) -> dict:
        """
        Turn on a device with ``switch`` capability.

        :param device: Device name or ID.
        :return: Device status
        """
        self.execute(device, 'switch', 'on')
        # noinspection PyUnresolvedReferences
        return self.status(device).output[0]

    @action
    def off(self, device: str, *args, **kwargs) -> dict:
        """
        Turn off a device with ``switch`` capability.

        :param device: Device name or ID.
        :return: Device status
        """
        self.execute(device, 'switch', 'off')
        # noinspection PyUnresolvedReferences
        return self.status(device).output[0]

    @action
    def toggle(self, device: str, *args, **kwargs) -> dict:
        """
        Toggle a device with ``switch`` capability.

        :param device: Device name or ID.
        :return: Device status
        """
        import pysmartthings

        device = self._get_device(device)
        device_id = device.device_id

        async def _toggle() -> bool:
            async with aiohttp.ClientSession(timeout=self._timeout) as session:
                api = pysmartthings.SmartThings(session, self._access_token)
                dev = await api.device(device_id)
                assert 'switch' in dev.capabilities, 'The device {} has no switch capability'.format(dev.label)

                await dev.status.refresh()
                state = 'off' if dev.status.switch else 'on'
                ret = await dev.command(component_id='main', capability='switch', command=state, args=args)

            assert ret, 'The command switch={state} failed on device {device}'.format(state=state, device=dev.label)
            return not dev.status.switch

        with self._refresh_lock:
            loop = asyncio.new_event_loop()
            state = loop.run_until_complete(_toggle())
            return {
                'id': device_id,
                'name': device.label,
                'on': state,
            }

    @property
    def switches(self) -> List[dict]:
        """
        :return: List of switch devices statuses in :class:`platypush.plugins.switch.SwitchPlugin` compatible format.
            Example:

          .. code-block:: json

            [
              {
                "id": "switch-1",
                "name": "Fan",
                "on": false
              },
              {
                "id": "tv-1",
                "name": "Samsung Smart TV",
                "on": true
              }
            ]

        """
        # noinspection PyUnresolvedReferences
        devices = self.status().output
        return [
            {
                'name': device['name'],
                'id': device['device_id'],
                'on': device['switch'],
            }
            for device in devices
            if 'switch' in device
        ]


# vim:sw=4:ts=4:et:



from datetime import datetime
from typing import List, Optional, Union, Iterable

from platypush.common.spotify import SpotifyMixin
from platypush.message.response import Response
from platypush.plugins import action
from platypush.plugins.media import PlayerState
from platypush.plugins.music import MusicPlugin
from platypush.schemas.spotify import (
    SpotifyDeviceSchema,
    SpotifyStatusSchema,
    SpotifyTrackSchema,
    SpotifyHistoryItemSchema,
    SpotifyPlaylistSchema,
    SpotifyAlbumSchema,
    SpotifyEpisodeSchema,
    SpotifyShowSchema,
    SpotifyArtistSchema,
)


class MusicSpotifyPlugin(MusicPlugin, SpotifyMixin):
    """
    Plugin to interact with the user's Spotify library and players.

    In order to use this plugin to interact with your Spotify account you need to register a new app on the Spotify
    developers website, whitelist the callback URL of your Platypush host and authorize the app to your account:

        - Create a developer app on https://developer.spotify.com.
        - Get the app's ``client_id`` and ``client_secret``.
        - Whitelist the authorization callback URL on the Platypush machine, usually in the form
          ``http(s)://your-platypush-hostname-or-local-ip:8008/spotify/auth_callback`` (you need the
          ``http`` Platypush backend to be enabled).
        - You can then authorize the app by opening the following URL in a browser:
          ``https://accounts.spotify.com/authorize?client_id=<client_id>&response_type=code&redirect_uri=http(s)://your-platypush-hostname-or-local-ip:8008/spotify/auth_callback&scope=<comma-separated-list-of-scopes>&state=<some-random-string>``.

    This is the list of scopes required for full plugin functionalities:

        - ``user-read-playback-state``
        - ``user-modify-playback-state``
        - ``user-read-currently-playing``
        - ``user-read-recently-played``
        - ``app-remote-control``
        - ``streaming``
        - ``playlist-modify-public``
        - ``playlist-modify-private``
        - ``playlist-read-private``
        - ``playlist-read-collaborative``
        - ``user-library-modify``
        - ``user-library-read``

    Alternatively, you can call any of the methods from this plugin over HTTP API, and the full authorization URL should
    be printed on the application logs/stdout.
    """

    def __init__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        **kwargs,
    ):
        MusicPlugin.__init__(self, **kwargs)
        SpotifyMixin.__init__(
            self, client_id=client_id, client_secret=client_secret, **kwargs
        )
        self._players_by_id = {}
        self._players_by_name = {}
        # Playlist ID -> snapshot ID and tracks cache
        self._playlist_snapshots = {}

    def _get_device(self, device: str):
        dev = self._players_by_id.get(device, self._players_by_name.get(device))
        if not dev:
            self.get_devices()

        dev = self._players_by_id.get(device, self._players_by_name.get(device))
        assert dev, f'No such device: {device}'
        return dev

    @staticmethod
    def _parse_datetime(
        dt: Optional[Union[str, datetime, int, float]]
    ) -> Optional[datetime]:
        if isinstance(dt, str):
            try:
                dt = float(dt)
            except (ValueError, TypeError):
                return datetime.fromisoformat(dt)

        if isinstance(dt, (int, float)):
            return datetime.fromtimestamp(dt)

        return dt

    @action
    def get_devices(self) -> List[dict]:
        """
        Get the list of players associated to the Spotify account.

        :return: .. schema:: spotify.SpotifyDeviceSchema(many=True)
        """
        devices = self.spotify_user_call('/v1/me/player/devices').get('devices', [])
        self._players_by_id = {
            **self._players_by_id,
            **{dev['id']: dev for dev in devices},
        }

        self._players_by_name = {
            **self._players_by_name,
            **{dev['name']: dev for dev in devices},
        }

        return SpotifyDeviceSchema().dump(devices, many=True)

    @action
    def set_volume(self, volume: int, device: Optional[str] = None):
        """
        Set the playback volume on a device.

        :param volume: Target volume as a percentage between 0 and 100.
        :param device: Device ID or name. If none is specified then the currently active device will be used.
        """
        if device:
            device = self._get_device(device)['id']

        self.spotify_user_call(
            '/v1/me/player/volume',
            method='put',
            params={
                'volume_percent': volume,
                **({'device_id': device} if device else {}),
            },
        )

    def _get_volume(self, device: Optional[str] = None) -> Optional[int]:
        if device:
            return self._get_device(device).get('volume')

        return self.status.output.get('volume')

    @action
    def volup(self, delta: int = 5, device: Optional[str] = None):
        """
        Set the volume up by a certain delta.

        :param delta: Increase the volume by this percentage amount (between 0 and 100).
        :param device: Device ID or name. If none is specified then the currently active device will be used.
        """
        if device:
            device = self._get_device(device)['id']

        self.spotify_user_call(
            '/v1/me/player/volume',
            params={
                'volume_percent': min(100, (self._get_volume() or 0) + delta),
                **({'device_id': device} if device else {}),
            },
        )

    @action
    def voldown(self, delta: int = 5, device: Optional[str] = None):
        """
        Set the volume down by a certain delta.

        :param delta: Decrease the volume by this percentage amount (between 0 and 100).
        :param device: Device ID or name. If none is specified then the currently active device will be used.
        """
        if device:
            device = self._get_device(device)['id']

        self.spotify_user_call(
            '/v1/me/player/volume',
            params={
                'volume_percent': max(0, (self._get_volume() or 0) - delta),
                **({'device_id': device} if device else {}),
            },
        )

    @action
    def play(self, resource: Optional[str] = None, device: Optional[str] = None):
        """
        Change the playback state of a device to ``PLAY`` or start playing a specific resource.

        :param resource: Resource to play, in Spotify URI format (e.g. ``spotify:track:xxxxxxxxxxxxxxxxxxxxxx``).
            If none is specified then the method will change the playback state to ``PLAY``.
        :param device: Device ID or name. If none is specified then the action will target the currently active device.
        """
        if device:
            device = self._get_device(device)['id']

        self.spotify_user_call(
            '/v1/me/player/play',
            method='put',
            json={'uris': [resource]} if resource else {},
            params={
                **({'device_id': device} if device else {}),
            },
        )

    @action
    def pause(self, device: Optional[str] = None):
        """
        Toggle paused state.

        :param device: Device ID or name. If none is specified then the action will target the currently active device.
        """
        if device:
            device = self._get_device(device)['id']

        # noinspection PyUnresolvedReferences
        status = self.status().output
        state = (
            'play'
            if status.get('device_id') != device
            or status.get('state') != PlayerState.PLAY.value
            else 'pause'
        )

        self.spotify_user_call(
            f'/v1/me/player/{state}',
            method='put',
            params={
                **({'device_id': device} if device else {}),
            },
        )

    @action
    def pause_if_playing(self):
        """
        Pause playback only if it's playing
        """
        # noinspection PyUnresolvedReferences
        status = self.status().output
        if status.get('state') == PlayerState.PLAY.value:
            self.spotify_user_call(
                '/v1/me/player/pause',
                method='put',
            )

    @action
    def play_if_paused(self, device: Optional[str] = None):
        """
        Play only if it's paused (resume)

        :param device: Device ID or name. If none is specified then the action will target the currently active device.
        """
        if device:
            device = self._get_device(device)['id']

        # noinspection PyUnresolvedReferences
        status = self.status().output
        if status.get('state') != PlayerState.PLAY.value:
            self.spotify_user_call(
                '/v1/me/player/play',
                method='put',
                params={
                    **({'device_id': device} if device else {}),
                },
            )

    @action
    def play_if_paused_or_stopped(self):
        """
        Alias for :meth:`.play_if_paused`.
        """
        return self.play_if_paused()

    @action
    def stop(self, **kwargs):
        """
        This method is actually just an alias to :meth:`.stop`, since Spotify manages clearing playback sessions
        automatically after a while for paused devices.
        """
        return self.pause(**kwargs)

    @action
    def start_or_transfer_playback(self, device: str):
        """
        Start or transfer playback to the device specified.

        :param device: Device ID or name.
        """
        device = self._get_device(device)['id']
        self.spotify_user_call(
            '/v1/me/player',
            method='put',
            json={
                'device_ids': [device],
            },
        )

    @action
    def next(self, device: Optional[str] = None, **kwargs):
        """
        Skip to the next track.

        :param device: Device ID or name. If none is specified then the action will target the currently active device.
        """
        if device:
            device = self._get_device(device)['id']

        self.spotify_user_call(
            '/v1/me/player/next',
            method='post',
            params={
                **({'device_id': device} if device else {}),
            },
        )

    @action
    def previous(self, device: Optional[str] = None, **kwargs):
        """
        Skip to the next track.

        :param device: Device ID or name. If none is specified then the action will target the currently active device.
        """
        if device:
            device = self._get_device(device)['id']

        self.spotify_user_call(
            '/v1/me/player/previous',
            method='post',
            params={
                **({'device_id': device} if device else {}),
            },
        )

    @action
    def seek(self, position: float, device: Optional[str] = None, **kwargs):
        """
        Set the cursor to the specified position in the track.

        :param position: Position in seconds.
        :param device: Device ID or name. If none is specified then the action will target the currently active device.
        """
        if device:
            device = self._get_device(device)['id']

        self.spotify_user_call(
            '/v1/me/player/seek',
            method='put',
            params={
                'position_ms': int(position * 1000),
                **({'device_id': device} if device else {}),
            },
        )

    @action
    def repeat(self, value: Optional[bool] = None, device: Optional[str] = None):
        """
        Set or toggle repeat mode.

        :param value: If set, set the repeat state this value (true/false). Default: None (toggle current state).
        :param device: Device ID or name. If none is specified then the action will target the currently active device.
        """
        if device:
            device = self._get_device(device)['id']

        if value is None:
            # noinspection PyUnresolvedReferences
            status = self.status().output
            state = (
                'context'
                if status.get('device_id') != device or not status.get('repeat')
                else 'off'
            )
        else:
            state = value is True

        self.spotify_user_call(
            '/v1/me/player/repeat',
            method='put',
            params={
                'state': 'context' if state else 'off',
                **({'device_id': device} if device else {}),
            },
        )

    @action
    def random(self, value: Optional[bool] = None, device: Optional[str] = None):
        """
        Set or toggle random/shuffle mode.

        :param value: If set, set the shuffle state this value (true/false). Default: None (toggle current state).
        :param device: Device ID or name. If none is specified then the action will target the currently active device.
        """
        if device:
            device = self._get_device(device)['id']

        if value is None:
            # noinspection PyUnresolvedReferences
            status = self.status().output
            state = bool(status.get('device_id') != device or not status.get('random'))
        else:
            state = value is True

        self.spotify_user_call(
            '/v1/me/player/shuffle',
            method='put',
            params={
                'state': state,
                **({'device_id': device} if device else {}),
            },
        )

    @action
    def history(
        self,
        limit: int = 20,
        before: Optional[Union[datetime, str, int]] = None,
        after: Optional[Union[datetime, str, int]] = None,
    ):
        """
        Get a list of recently played track on the account.

        :param limit: Maximum number of tracks to be retrieved (default: 20, max: 50).
        :param before: Retrieve only the tracks played before this timestamp, specified as a UNIX timestamp, a datetime
            object or an ISO datetime string. If ``before`` is set then ``after`` cannot be set.
        :param after: Retrieve only the tracks played after this timestamp, specified as a UNIX timestamp, a datetime
            object or an ISO datetime string. If ``after`` is set then ``before`` cannot be set.
        :return:
        """
        before = self._parse_datetime(before)
        after = self._parse_datetime(after)
        assert not (before and after), 'before and after cannot both be set'

        results = self._spotify_paginate_results(
            '/v1/me/player/recently-played',
            limit=limit,
            params={
                'limit': min(limit, 50),
                **({'before': before} if before else {}),
                **({'after': after} if after else {}),
            },
        )

        return SpotifyHistoryItemSchema().dump(
            [
                {
                    **item.pop('track'),
                    **item,
                }
                for item in results
            ],
            many=True,
        )

    @action
    def add(self, resource: str, device: Optional[str] = None, **kwargs):
        """
        Add a Spotify resource (track, or episode) to the playing queue.

        :param resource: Spotify resource URI.
        :param device: Device ID or name. If none is specified then the action will target the currently active device.
        """
        if device:
            device = self._get_device(device)['id']

        self.spotify_user_call(
            '/v1/me/player/queue',
            method='post',
            params={
                'uri': resource,
                **({'device_id': device} if device else {}),
            },
        )

    @action
    def clear(self, **kwargs):
        pass

    @action
    def status(self, **kwargs) -> dict:
        """
        Get the status of the currently active player.

        :return: .. schema:: spotify.SpotifyStatusSchema
        """
        status = self.spotify_user_call('/v1/me/player')
        if not status:
            return {
                'state': PlayerState.STOP.value,
            }

        return SpotifyStatusSchema().dump(status)

    @action
    def current_track(self, **kwargs) -> dict:
        """
        Get the track currently playing.

        :return: .. schema:: spotify.SpotifyTrackSchema
        """
        status = self.spotify_user_call('/v1/me/player')
        empty_response = Response(output={})
        if not status:
            # noinspection PyTypeChecker
            return empty_response

        track = status.get('item', {})
        if not track:
            # noinspection PyTypeChecker
            return empty_response

        return SpotifyTrackSchema().dump(track)

    @action
    def get_playlists(
        self, limit: int = 1000, offset: int = 0, user: Optional[str] = None
    ):
        """
        Get the user's playlists.

        :param limit: Maximum number of results (default: 1000).
        :param offset: Return results starting from this index (default: 0).
        :param user: Return the playlist owned by a specific user ID (default: currently logged in user).
        :return: .. schema:: spotify.SpotifyPlaylistSchema
        """
        playlists = self._spotify_paginate_results(
            f'/v1/{"users/" + user if user else "me"}/playlists',
            limit=limit,
            offset=offset,
        )

        return SpotifyPlaylistSchema().dump(playlists, many=True)

    def _get_playlist(self, playlist: str) -> dict:
        playlists = self.get_playlists().output
        playlists = [
            pl
            for pl in playlists
            if (pl['id'] == playlist or pl['uri'] == playlist or pl['name'] == playlist)
        ]

        assert playlists, f'No such playlist ID, URI or name: {playlist}'
        return playlists[0]

    def _get_playlist_tracks_from_cache(
        self, id: str, snapshot_id: str, limit: Optional[int] = None, offset: int = 0
    ) -> Optional[Iterable]:
        snapshot = self._playlist_snapshots.get(id)
        if (
            not snapshot
            or snapshot['snapshot_id'] != snapshot_id
            or (limit is None and snapshot['limit'] is not None)
        ):
            return

        if limit is not None and snapshot['limit'] is not None:
            stored_range = (snapshot['limit'], snapshot['limit'] + snapshot['offset'])
            requested_range = (limit, limit + offset)
            if (
                requested_range[0] < stored_range[0]
                or requested_range[1] > stored_range[1]
            ):
                return

        return snapshot['tracks']

    def _cache_playlist_data(
        self,
        id: str,
        snapshot_id: str,
        tracks: Iterable[dict],
        limit: Optional[int] = None,
        offset: int = 0,
        **_,
    ):
        self._playlist_snapshots[id] = {
            'id': id,
            'tracks': tracks,
            'snapshot_id': snapshot_id,
            'limit': limit,
            'offset': offset,
        }

    @action
    def get_playlist(
        self,
        playlist: str,
        with_tracks: bool = True,
        limit: Optional[int] = None,
        offset: int = 0,
    ):
        """
        Get a playlist content.

        :param playlist: Playlist name, ID or URI.
        :param with_tracks: Return also the playlist tracks (default: false, return only the metadata).
        :param limit: If ``with_tracks`` is True, retrieve this maximum amount of tracks
            (default: None, get all tracks).
        :param offset: If ``with_tracks`` is True, retrieve tracks starting from this index (default: 0).
        :return: .. schema:: spotify.SpotifyPlaylistSchema
        """
        playlist = self._get_playlist(playlist)
        if with_tracks:
            playlist['tracks'] = self._get_playlist_tracks_from_cache(
                playlist['id'],
                snapshot_id=playlist['snapshot_id'],
                limit=limit,
                offset=offset,
            )

            if playlist['tracks'] is None:
                playlist['tracks'] = [
                    {
                        **track,
                        'track': {
                            **track['track'],
                            'position': offset + i + 1,
                        },
                    }
                    for i, track in enumerate(
                        self._spotify_paginate_results(
                            f'/v1/playlists/{playlist["id"]}/tracks',
                            limit=limit,
                            offset=offset,
                        )
                    )
                ]

                self._cache_playlist_data(**playlist, limit=limit, offset=offset)

        return SpotifyPlaylistSchema().dump(playlist)

    @action
    def add_to_playlist(
        self,
        playlist: str,
        resources: Union[str, Iterable[str]],
        position: Optional[int] = None,
    ):
        """
        Add one or more items to a playlist.

        :param playlist: Playlist name, ID or URI.
        :param resources: URI(s) of the resource(s) to be added.
        :param position: At what (1-based) position the tracks should be inserted (default: append to the end).
        """
        playlist = self._get_playlist(playlist)
        response = self.spotify_user_call(
            f'/v1/playlists/{playlist["id"]}/tracks',
            method='post',
            params={
                **({'position': position} if position is not None else {}),
            },
            json={
                'uris': [
                    uri.strip()
                    for uri in (
                        resources.split(',')
                        if isinstance(resources, str)
                        else resources
                    )
                ]
            },
        )

        snapshot_id = response.get('snapshot_id')
        assert snapshot_id is not None, 'Could not save playlist'

    @action
    def remove_from_playlist(self, playlist: str, resources: Union[str, Iterable[str]]):
        """
        Remove one or more items from a playlist.

        :param playlist: Playlist name, ID or URI.
        :param resources: URI(s) of the resource(s) to be removed. A maximum of 100 tracks can be provided at once.
        """
        playlist = self._get_playlist(playlist)
        response = self.spotify_user_call(
            f'/v1/playlists/{playlist["id"]}/tracks',
            method='delete',
            json={
                'tracks': [
                    {'uri': uri.strip()}
                    for uri in (
                        resources.split(',')
                        if isinstance(resources, str)
                        else resources
                    )
                ]
            },
        )

        snapshot_id = response.get('snapshot_id')
        assert snapshot_id is not None, 'Could not save playlist'

    @action
    def playlist_move(
        self,
        playlist: str,
        from_pos: int,
        to_pos: int,
        range_length: int = 1,
        resources: Optional[Union[str, Iterable[str]]] = None,
        **_,
    ):
        """
        Move or replace elements in a playlist.

        :param playlist: Playlist name, ID or URI.
        :param from_pos: Move tracks starting from this position (the first element has index 1).
        :param to_pos: Move tracks to this position (1-based index).
        :param range_length: Number of tracks to move (default: 1).
        :param resources: If specified, then replace the items from `from_pos` to `from_pos+range_length` with the
            specified set of Spotify URIs (it must be a collection with the same length as the range).
        """
        playlist = self._get_playlist(playlist)
        response = self.spotify_user_call(
            f'/v1/playlists/{playlist["id"]}/tracks',
            method='put',
            json={
                'range_start': int(from_pos) + 1,
                'range_length': int(range_length),
                'insert_before': int(to_pos) + 1,
                **(
                    {
                        'uris': [
                            uri.strip()
                            for uri in (
                                resources.split(',')
                                if isinstance(resources, str)
                                else resources
                            )
                        ]
                    }
                    if resources
                    else {}
                ),
            },
        )

        snapshot_id = response.get('snapshot_id')
        assert snapshot_id is not None, 'Could not save playlist'

    # noinspection PyShadowingBuiltins
    @staticmethod
    def _make_filter(query: Union[str, dict], **filter) -> str:
        if filter:
            query = {
                **({'any': query} if isinstance(query, str) else {}),
                **filter,
            }

        if isinstance(query, str):
            return query

        q = query['any'] if 'any' in query else ''
        for attr in ['artist', 'track', 'album', 'year']:
            if attr in query:
                q += f' {attr}:{query[attr]}'

        return q.strip()

    # noinspection PyShadowingBuiltins
    @action
    def search(
        self,
        query: Optional[Union[str, dict]] = None,
        limit: int = 50,
        offset: int = 0,
        type: str = 'track',
        **filter,
    ) -> Iterable[dict]:
        """
        Search for tracks matching a certain criteria.

        :param query: Search filter. It can either be a free-text or a structured query. In the latter case the
            following fields are supported:

                - ``any``: Search for anything that matches this text.
                - ``uri``: Search the following Spotify ID/URI or list of IDs/URIs.
                - ``artist``: Filter by artist.
                - ``track``: Filter by track name.
                - ``album``: Filter by album name.
                - ``year``: Filter by year (dash-separated ranges are supported).

        :param limit: Maximum number of results (default: 50).
        :param offset: Return results starting from this index (default: 0).
        :param type: Type of results to be returned. Supported: ``album``, ``artist``, ``playlist``, ``track``, ``show``
            and ``episode`` (default: ``track``).
        :param filter: Alternative key-value way of representing a structured query.
        :return:
            If ``type=track``:
                .. schema:: spotify.SpotifyTrackSchema(many=True)
            If ``type=album``:
                .. schema:: spotify.SpotifyAlbumSchema(many=True)
            If ``type=artist``:
                .. schema:: spotify.SpotifyArtistSchema(many=True)
            If ``type=playlist``:
                .. schema:: spotify.SpotifyPlaylistSchema(many=True)
            If ``type=episode``:
                .. schema:: spotify.SpotifyEpisodeSchema(many=True)
            If ``type=show``:
                .. schema:: spotify.SpotifyShowSchema(many=True)

        """
        uri = {
            **(query if isinstance(query, dict) else {}),
            **filter,
        }.get('uri', [])

        uris = uri.split(',') if isinstance(uri, str) else uri
        params = (
            {
                'ids': ','.join([uri.split(':')[-1].strip() for uri in uris]),
            }
            if uris
            else {
                'q': self._make_filter(query, **filter),
                'type': type,
            }
        )

        response = self._spotify_paginate_results(
            f'/v1/{type + "s" if uris else "search"}',
            limit=limit,
            offset=offset,
            type=type,
            params=params,
        )

        if type == 'track':
            return sorted(
                SpotifyTrackSchema(many=True).dump(response),
                key=lambda track: (
                    track.get('artist'),
                    track.get('date'),
                    track.get('album'),
                    track.get('track'),
                    track.get('title'),
                    track.get('popularity'),
                ),
            )

        schema_class = None
        if type == 'playlist':
            schema_class = SpotifyPlaylistSchema
        if type == 'album':
            schema_class = SpotifyAlbumSchema
        if type == 'artist':
            schema_class = SpotifyArtistSchema
        if type == 'episode':
            schema_class = SpotifyEpisodeSchema
        if type == 'show':
            schema_class = SpotifyShowSchema

        if schema_class:
            return schema_class(many=True).dump(response)

        return response

    @action
    def create_playlist(
        self, name: str, description: Optional[str] = None, public: bool = False
    ):
        """
        Create a playlist.

        :param name: Playlist name.
        :param description: Optional playlist description.
        :param public: Whether the new playlist should be public
            (default: False).
        :return: .. schema:: spotify.SpotifyPlaylistSchema
        """
        ret = self.spotify_user_call(
            '/v1/users/me/playlists',
            method='post',
            json={
                'name': name,
                'description': description,
                'public': public,
            },
        )

        return SpotifyPlaylistSchema().dump(ret)

    @action
    def follow_playlist(self, playlist: str, public: bool = True):
        """
        Follow a playlist.

        :param playlist: Playlist name, ID or URI.
        :param public: If True (default) then the playlist will appear in the user's list of public playlists, otherwise
            it won't.
        """
        playlist = self._get_playlist(playlist)
        self.spotify_user_call(
            f'/v1/playlists/{playlist["id"]}/followers',
            method='put',
            json={
                'public': public,
            },
        )

    @action
    def unfollow_playlist(self, playlist: str):
        """
        Unfollow a playlist.

        :param playlist: Playlist name, ID or URI.
        """
        playlist = self._get_playlist(playlist)
        self.spotify_user_call(
            f'/v1/playlists/{playlist["id"]}/followers',
            method='delete',
        )

    @staticmethod
    def _uris_to_id(*uris: str) -> Iterable[str]:
        return [uri.split(':')[-1] for uri in uris]

    @action
    def get_albums(self, limit: int = 50, offset: int = 0) -> List[dict]:
        """
        Get the list of albums saved by the user.

        :param limit: Maximum number of results (default: 50).
        :param offset: Return results starting from this index (default: 0).
        :return: .. schema:: spotify.SpotifyAlbumSchema(many=True)
        """
        return SpotifyAlbumSchema().dump(
            self._spotify_paginate_results(
                '/v1/me/albums',
                limit=limit,
                offset=offset,
            ),
            many=True,
        )

    @action
    def save_albums(self, resources: Iterable[str]):
        """
        Save a list of albums to the user's collection.

        :param resources: Spotify IDs or URIs of the albums to save.
        """
        self.spotify_user_call(
            '/v1/me/albums',
            method='put',
            json={'ids': self._uris_to_id(*resources)},
        )

    @action
    def remove_albums(self, resources: Iterable[str]):
        """
        Remove a list of albums from the user's collection.

        :param resources: Spotify IDs or URIs of the albums to remove.
        """
        self.spotify_user_call(
            '/v1/me/albums',
            method='delete',
            json={'ids': self._uris_to_id(*resources)},
        )

    @action
    def get_tracks(self, limit: int = 100, offset: int = 0) -> List[dict]:
        """
        Get the list of tracks saved by the user.

        :param limit: Maximum number of results (default: 100).
        :param offset: Return results starting from this index (default: 0).
        :return: .. schema:: spotify.SpotifyTrackSchema(many=True)
        """
        return [
            SpotifyTrackSchema().dump(item['track'])
            for item in self._spotify_paginate_results(
                '/v1/me/tracks', limit=limit, offset=offset
            )
        ]

    @action
    def save_tracks(self, resources: Iterable[str]):
        """
        Save a list of tracks to the user's collection.

        :param resources: Spotify IDs or URIs of the tracks to save.
        """
        self.spotify_user_call(
            '/v1/me/tracks',
            method='put',
            json={'ids': self._uris_to_id(*resources)},
        )

    @action
    def remove_tracks(self, resources: Iterable[str]):
        """
        Remove a list of tracks from the user's collection.

        :param resources: Spotify IDs or URIs of the tracks to remove.
        """
        self.spotify_user_call(
            '/v1/me/tracks',
            method='delete',
            json={'ids': self._uris_to_id(*resources)},
        )

    @action
    def get_episodes(self, limit: int = 50, offset: int = 0) -> List[dict]:
        """
        Get the list of episodes saved by the user.

        :param limit: Maximum number of results (default: 50).
        :param offset: Return results starting from this index (default: 0).
        :return: .. schema:: spotify.SpotifyEpisodeSchema(many=True)
        """
        return SpotifyEpisodeSchema().dump(
            self._spotify_paginate_results(
                '/v1/me/episodes',
                limit=limit,
                offset=offset,
            ),
            many=True,
        )

    @action
    def save_episodes(self, resources: Iterable[str]):
        """
        Save a list of episodes to the user's collection.

        :param resources: Spotify IDs or URIs of the episodes to save.
        """
        self.spotify_user_call(
            '/v1/me/episodes',
            method='put',
            json={'ids': self._uris_to_id(*resources)},
        )

    @action
    def remove_episodes(self, resources: Iterable[str]):
        """
        Remove a list of episodes from the user's collection.

        :param resources: Spotify IDs or URIs of the episodes to remove.
        """
        self.spotify_user_call(
            '/v1/me/episodes',
            method='delete',
            json={'ids': self._uris_to_id(*resources)},
        )

    @action
    def get_shows(self, limit: int = 50, offset: int = 0) -> List[dict]:
        """
        Get the list of shows saved by the user.

        :param limit: Maximum number of results (default: 50).
        :param offset: Return results starting from this index (default: 0).
        :return: .. schema:: spotify.SpotifyShowSchema(many=True)
        """
        return SpotifyShowSchema().dump(
            self._spotify_paginate_results(
                '/v1/me/shows',
                limit=limit,
                offset=offset,
            ),
            many=True,
        )

    @action
    def save_shows(self, resources: Iterable[str]):
        """
        Save a list of shows to the user's collection.

        :param resources: Spotify IDs or URIs of the shows to save.
        """
        self.spotify_user_call(
            '/v1/me/shows',
            method='put',
            json={'ids': self._uris_to_id(*resources)},
        )

    @action
    def remove_shows(self, resources: Iterable[str]):
        """
        Remove a list of shows from the user's collection.

        :param resources: Spotify IDs or URIs of the shows to remove.
        """
        self.spotify_user_call(
            '/v1/me/shows',
            method='delete',
            json={'ids': self._uris_to_id(*resources)},
        )



import enum
import inspect
import json
import logging
import os
import pathlib
import shutil
import yaml

from abc import ABC, abstractmethod
from typing import Optional, Iterable, Mapping, Callable, Type

supported_package_managers = {
    'pacman': 'pacman -S',
    'apt': 'apt-get install',
}

_available_package_manager = None


class ManifestType(enum.Enum):
    PLUGIN = 'plugin'
    BACKEND = 'backend'


class Manifest(ABC):
    """
    Base class for plugin/backend manifests.
    """
    def __init__(self, package: str, description: Optional[str] = None,
                 install: Optional[Iterable[str]] = None, events: Optional[Mapping] = None, **_):
        self.description = description
        self.install = install or {}
        self.events = events or {}
        self.logger = logging.getLogger(__name__)
        self.package = package
        self.component_name = '.'.join(package.split('.')[2:])
        self.component = None

    @classmethod
    @property
    @abstractmethod
    def component_getter(self) -> Callable[[str], object]:
        raise NotImplementedError

    @classmethod
    def from_file(cls, filename: str) -> "Manifest":
        with open(str(filename), 'r') as f:
            manifest = yaml.safe_load(f).get('manifest', {})

        assert 'type' in manifest, f'Manifest file {filename} has no type field'
        comp_type = ManifestType(manifest.pop('type'))
        manifest_class = _manifest_class_by_type[comp_type]
        return manifest_class(**manifest)

    @classmethod
    def from_class(cls, clazz) -> "Manifest":
        return cls.from_file(os.path.dirname(inspect.getfile(clazz)))

    @classmethod
    def from_component(cls, comp) -> "Manifest":
        return cls.from_class(comp.__class__)

    def get_component(self):
        try:
            self.component = self.component_getter(self.component_name)
        except Exception as e:
            self.logger.warning(f'Could not load {self.component_name}: {e}')

        return self.component

    def __repr__(self):
        return json.dumps({
            'description': self.description,
            'install': self.install,
            'events': self.events,
            'type': _manifest_type_by_class[self.__class__].value,
            'package': self.package,
            'component_name': self.component_name,
        })


class PluginManifest(Manifest):
    @classmethod
    @property
    def component_getter(self):
        from platypush.context import get_plugin
        return get_plugin


class BackendManifest(Manifest):
    @classmethod
    @property
    def component_getter(self):
        from platypush.context import get_backend
        return get_backend


_manifest_class_by_type: Mapping[ManifestType, Type[Manifest]] = {
    ManifestType.PLUGIN: PluginManifest,
    ManifestType.BACKEND: BackendManifest,
}

_manifest_type_by_class: Mapping[Type[Manifest], ManifestType] = {
    cls: t for t, cls in _manifest_class_by_type.items()
}


def scan_manifests(base_class: Type) -> Iterable[str]:
    for mf in pathlib.Path(os.path.dirname(inspect.getfile(base_class))).rglob('manifest.yaml'):
        yield str(mf)


def get_manifests(base_class: Type) -> Iterable[Manifest]:
    return [
        Manifest.from_file(mf)
        for mf in scan_manifests(base_class)
    ]


def get_components(base_class: Type) -> Iterable:
    manifests = get_manifests(base_class)
    components = {mf.get_component() for mf in manifests}
    return {comp for comp in components if comp is not None}


def get_manifests_from_conf(conf_file: Optional[str] = None) -> Mapping[str, Manifest]:
    import platypush
    from platypush.config import Config

    conf_args = []
    if conf_file:
        conf_args.append(conf_file)

    Config.init(*conf_args)
    app_dir = os.path.dirname(inspect.getfile(platypush))
    manifest_files = set()

    for name in Config.get_backends().keys():
        manifest_files.add(os.path.join(app_dir, 'backend', *name.split('.'), 'manifest.yaml'))

    for name in Config.get_plugins().keys():
        manifest_files.add(os.path.join(app_dir, 'plugins', *name.split('.'), 'manifest.yaml'))

    return {
        manifest_file: Manifest.from_file(manifest_file)
        for manifest_file in manifest_files
    }


def get_dependencies_from_conf(conf_file: Optional[str] = None) -> Mapping[str, Iterable[str]]:
    manifests = get_manifests_from_conf(conf_file)
    deps = {
        'pip': set(),
        'packages': set(),
        'exec': set(),
    }

    for manifest in manifests.values():
        deps['pip'].update(manifest.install.get('pip', set()))
        deps['exec'].update(manifest.install.get('exec', set()))
        has_requires_packages = len([
            section for section in manifest.install.keys()
            if section in supported_package_managers
        ]) > 0

        if has_requires_packages:
            pkg_manager = get_available_package_manager()
            deps['packages'].update(manifest.install.get(pkg_manager, set()))

    return deps


def get_install_commands_from_conf(conf_file: Optional[str] = None) -> Mapping[str, str]:
    deps = get_dependencies_from_conf(conf_file)
    return {
        'pip': f'pip install {" ".join(deps["pip"])}',
        'exec': deps["exec"],
        'packages': f'{supported_package_managers[_available_package_manager]} {" ".join(deps["packages"])}'
        if deps['packages'] else None,
    }


def get_available_package_manager() -> str:
    global _available_package_manager
    if _available_package_manager:
        return _available_package_manager

    available_package_managers = [
        pkg_manager for pkg_manager in supported_package_managers.keys()
        if shutil.which(pkg_manager)
    ]

    assert available_package_managers, (
        'Your OS does not provide any of the supported package managers. '
        f'Supported package managers: {supported_package_managers.keys()}'
    )

    _available_package_manager = available_package_managers[0]
    return _available_package_manager



import threading
import time

from typing import Optional, List, Tuple, Union

from platypush.plugins import action
from platypush.plugins.camera import CameraPlugin, Camera
from platypush.plugins.camera.pi.model import PiCameraInfo, PiCamera


class CameraPiPlugin(CameraPlugin):
    """
    Plugin to control a Pi camera.

    Requires:

        * **picamera** (``pip install picamera``)
        * **numpy** (``pip install numpy``)
        * **Pillow** (``pip install Pillow``)

    """

    _camera_class = PiCamera
    _camera_info_class = PiCameraInfo

    def __init__(self, device: int = 0, fps: float = 30., warmup_seconds: float = 2., sharpness: int = 0,
                 contrast: int = 0, brightness: int = 50, video_stabilization: bool = False, iso: int = 0,
                 exposure_compensation: int = 0, exposure_mode: str = 'auto', meter_mode: str = 'average',
                 awb_mode: str = 'auto', image_effect: str = 'none', led_pin: Optional[int] = None,
                 color_effects: Optional[Union[str, List[str]]] = None,
                 zoom: Tuple[float, float, float, float] = (0.0, 0.0, 1.0, 1.0), **camera):
        """
        See https://www.raspberrypi.org/documentation/usage/camera/python/README.md
        for a detailed reference about the Pi camera options.

        :param camera: Options for the base camera plugin (see :class:`platypush.plugins.camera.CameraPlugin`).
        """
        super().__init__(device=device, fps=fps, warmup_seconds=warmup_seconds, **camera)

        self.camera_info.sharpness = sharpness
        self.camera_info.contrast = contrast
        self.camera_info.brightness = brightness
        self.camera_info.video_stabilization = video_stabilization
        self.camera_info.iso = iso
        self.camera_info.exposure_compensation = exposure_compensation
        self.camera_info.meter_mode = meter_mode
        self.camera_info.exposure_mode = exposure_mode
        self.camera_info.awb_mode = awb_mode
        self.camera_info.image_effect = image_effect
        self.camera_info.color_effects = color_effects
        self.camera_info.zoom = zoom
        self.camera_info.led_pin = led_pin

    # noinspection DuplicatedCode
    def prepare_device(self, device: PiCamera):
        # noinspection PyUnresolvedReferences
        import picamera

        camera = picamera.PiCamera(camera_num=device.info.device, resolution=device.info.resolution,
                                   framerate=device.info.fps, led_pin=device.info.led_pin)

        camera.hflip = device.info.horizontal_flip
        camera.vflip = device.info.vertical_flip
        camera.sharpness = device.info.sharpness
        camera.contrast = device.info.contrast
        camera.brightness = device.info.brightness
        camera.video_stabilization = device.info.video_stabilization
        camera.iso = device.info.iso
        camera.exposure_compensation = device.info.exposure_compensation
        camera.exposure_mode = device.info.exposure_mode
        camera.meter_mode = device.info.meter_mode
        camera.awb_mode = device.info.awb_mode
        camera.image_effect = device.info.image_effect
        camera.color_effects = device.info.color_effects
        camera.rotation = device.info.rotate or 0
        camera.zoom = device.info.zoom

        return camera

    def release_device(self, device: PiCamera):
        # noinspection PyUnresolvedReferences
        import picamera

        if device.object:
            try:
                device.object.stop_recording()
            except (ConnectionError, picamera.PiCameraNotRecording):
                pass

        if device.object and not device.object.closed:
            try:
                device.object.close()
            except (ConnectionError, picamera.PiCameraClosed):
                pass

    def capture_frame(self, camera: Camera, *args, **kwargs):
        import numpy as np
        from PIL import Image

        shape = (camera.info.resolution[1] + (camera.info.resolution[1] % 16),
                 camera.info.resolution[0] + (camera.info.resolution[0] % 32),
                 3)

        frame = np.empty(shape, dtype=np.uint8)
        camera.object.capture(frame, 'rgb')
        return Image.fromarray(frame)

    def start_preview(self, camera: Camera):
        """
        Start camera preview.
        """
        camera.object.start_preview()

    def stop_preview(self, camera: Camera):
        """
        Stop camera preview.
        """
        try:
            camera.object.stop_preview()
        except Exception as e:
            self.logger.warning(str(e))

    @action
    def capture_preview(self, duration: Optional[float] = None, n_frames: Optional[int] = None, **camera) -> dict:
        camera = self.open_device(**camera)
        self.start_preview(camera)

        if n_frames:
            duration = n_frames * (camera.info.fps or 0)
        if duration:
            threading.Timer(duration, lambda: self.stop_preview(camera))

        return self.status()

    def streaming_thread(self, camera: PiCamera, stream_format: str, duration: Optional[float] = None):
        server_socket = self._prepare_server_socket(camera)
        sock = None
        streaming_started_time = time.time()
        self.logger.info('Starting streaming on port {}'.format(camera.info.listen_port))

        try:
            while camera.stream_event.is_set():
                if duration and time.time() - streaming_started_time >= duration:
                    break

                sock = self._accept_client(server_socket)
                if not sock:
                    continue

                if camera.object is None or camera.object.closed:
                    camera = self.open_device(**camera.info.to_dict())

                try:
                    camera.object.start_recording(sock, format=stream_format)
                    while camera.stream_event.is_set():
                        camera.object.wait_recording(1)
                except ConnectionError:
                    self.logger.info('Client closed connection')
                finally:
                    if sock:
                        try:
                            sock.close()
                        except Exception as e:
                            self.logger.warning('Error while closing client socket: {}'.format(str(e)))

                    self.close_device(camera)
        finally:
            self._cleanup_stream(camera, server_socket, sock)
            self.logger.info('Stopped camera stream')

    @action
    def start_streaming(self, duration: Optional[float] = None, stream_format: str = 'h264', **camera) -> dict:
        camera = self.open_device(stream_format=stream_format, **camera)
        return self._start_streaming(camera, duration, stream_format)


# vim:sw=4:ts=4:et:



import copy
import json
import random
import re
import sys
import time

from datetime import date

from platypush.config import Config
from platypush.message import Message
from platypush.utils import get_event_class_by_type


class Event(Message):
    """Event message class"""

    # If this class property is set to false then the logging of these events
    # will be disabled. Logging is usually disabled for events with a very
    # high frequency that would otherwise pollute the logs e.g. camera capture
    # events
    # pylint: disable=redefined-builtin
    def __init__(
        self,
        target=None,
        origin=None,
        id=None,
        timestamp=None,
        disable_logging=False,
        disable_web_clients_notification=False,
        **kwargs
    ):
        """
        Params:
            target  -- Target node [String]
            origin  -- Origin node (default: current node) [String]
            id      -- Event ID (default: auto-generated)
            kwargs  -- Additional arguments for the event [kwDict]
        """

        super().__init__(timestamp=timestamp)
        self.id = id if id else self._generate_id()
        self.target = target if target else Config.get('device_id')
        self.origin = origin if origin else Config.get('device_id')
        self.type = '{}.{}'.format(self.__class__.__module__, self.__class__.__name__)
        self.args = kwargs
        self.disable_logging = disable_logging
        self.disable_web_clients_notification = disable_web_clients_notification

        for arg, value in self.args.items():
            if arg not in [
                'id',
                'args',
                'origin',
                'target',
                'type',
                'timestamp',
                'disable_logging',
            ] and not arg.startswith('_'):
                self.__setattr__(arg, value)

    @classmethod
    def build(cls, msg):
        """Builds an event message from a JSON UTF-8 string/bytearray, a
        dictionary, or another Event"""

        msg = super().parse(msg)
        event_type = msg['args'].pop('type')
        event_class = get_event_class_by_type(event_type)

        args = msg['args'] if 'args' in msg else {}
        args['id'] = msg['id'] if 'id' in msg else cls._generate_id()
        args['target'] = msg['target'] if 'target' in msg else Config.get('device_id')
        args['origin'] = msg['origin'] if 'origin' in msg else Config.get('device_id')
        args['timestamp'] = msg['_timestamp'] if '_timestamp' in msg else time.time()
        return event_class(**args)

    @staticmethod
    def _generate_id():
        """Generate a unique event ID"""
        return ''.join(['{:02x}'.format(random.randint(0, 255)) for _ in range(16)])

    def matches_condition(self, condition):
        """
        If the event matches an event condition, it will return an EventMatchResult
        :param condition: The platypush.event.hook.EventCondition object
        """

        result = EventMatchResult(is_match=False, parsed_args=self.args)
        match_scores = []

        if not isinstance(self, condition.type):
            return result

        for (attr, value) in condition.args.items():
            if attr not in self.args:
                return result

            if isinstance(self.args[attr], str):
                arg_result = self._matches_argument(argname=attr, condition_value=value)

                if arg_result.is_match:
                    match_scores.append(arg_result.score)
                    for (parsed_arg, parsed_value) in arg_result.parsed_args.items():
                        result.parsed_args[parsed_arg] = parsed_value
                else:
                    return result
            elif self.args[attr] != value:
                # TODO proper support for list and dictionary matches
                return result

        result.is_match = True
        if match_scores:
            result.score = sum(match_scores) / float(len(match_scores))

        return result

    def _matches_argument(self, argname, condition_value):
        """
        Returns an EventMatchResult if the event argument [argname] matches
        [condition_value].

        - Example:
            - self.args = {
                'phrase': 'Hey dude turn on the living room lights'
            }

            - self._matches_argument(argname='phrase', condition_value='Turn on the ${lights} lights')
              will return EventMatchResult(is_match=True, parsed_args={ 'lights': 'living room' })

            - self._matches_argument(argname='phrase', condition_value='Turn off the ${lights} lights')
              will return EventMatchResult(is_match=False, parsed_args={})
        """

        result = EventMatchResult(is_match=False)
        if self.args.get(argname) == condition_value:
            # In case of an exact match, return immediately
            result.is_match = True
            result.score = sys.maxsize
            return result

        event_tokens = re.split(r'\s+', self.args.get(argname, '').strip().lower())
        condition_tokens = re.split(r'\s+', condition_value.strip().lower())

        while event_tokens and condition_tokens:
            event_token = event_tokens[0]
            condition_token = condition_tokens[0]

            if event_token == condition_token:
                event_tokens.pop(0)
                condition_tokens.pop(0)
                result.score += 1.5
            elif re.search(condition_token, event_token):
                m = re.search('({})'.format(condition_token), event_token)
                if m.group(1):
                    event_tokens.pop(0)
                    result.score += 1.25

                condition_tokens.pop(0)
            else:
                m = re.match(r'[^\\]*\${(.+?)}', condition_token)
                if m:
                    argname = m.group(1)
                    if argname not in result.parsed_args:
                        result.parsed_args[argname] = event_token
                        result.score += 1.0
                    else:
                        result.parsed_args[argname] += ' ' + event_token

                    if (len(condition_tokens) == 1 and len(event_tokens) == 1) or (
                        len(event_tokens) > 1
                        and len(condition_tokens) > 1
                        and event_tokens[1] == condition_tokens[1]
                    ):
                        # Stop appending tokens to this argument, as the next
                        # condition will be satisfied as well
                        condition_tokens.pop(0)

                    event_tokens.pop(0)
                else:
                    result.score -= 1.0
                    event_tokens.pop(0)

        # It's a match if all the tokens in the condition string have been satisfied
        result.is_match = len(condition_tokens) == 0
        return result

    def __str__(self):
        """
        Overrides the str() operator and converts
        the message into a UTF-8 JSON string
        """

        args = copy.deepcopy(self.args)
        flatten(args)

        return json.dumps(
            {
                'type': 'event',
                'target': self.target,
                'origin': self.origin if hasattr(self, 'origin') else None,
                'id': self.id if hasattr(self, 'id') else None,
                '_timestamp': self.timestamp,
                'args': {'type': self.type, **args},
            },
            cls=self.Encoder,
        )


class EventMatchResult:
    """When comparing an event against an event condition, you want to
    return this object. It contains the match status (True or False),
    any parsed arguments, and a match_score that identifies how "strong"
    the match is - in case of multiple event matches, the ones with the
    highest score will win"""

    def __init__(self, is_match, score=0, parsed_args=None):
        self.is_match = is_match
        self.score = score
        self.parsed_args = parsed_args or {}


def flatten(args):
    if isinstance(args, dict):
        for (key, value) in args.items():
            if isinstance(value, date):
                args[key] = value.isoformat()
            elif isinstance(value, (dict, list)):
                flatten(args[key])
    elif isinstance(args, list):
        for i, arg in enumerate(args):
            if isinstance(arg, date):
                args[i] = arg.isoformat()
            elif isinstance(arg, (dict, list)):
                flatten(args[i])


# vim:sw=4:ts=4:et:



import datetime
import enum
import logging
import threading

import croniter
from dateutil.tz import gettz

from platypush.procedure import Procedure
from platypush.utils import is_functional_cron, set_thread_name

logger = logging.getLogger('platypush:cron')


def get_now() -> datetime.datetime:
    """
    :return: A timezone-aware representation of `now`
    """
    return datetime.datetime.now().replace(
        tzinfo=gettz()  # lgtm [py/call-to-non-callable]
    )


class CronjobState(enum.IntEnum):
    IDLE = 0
    WAIT = 1
    RUNNING = 2
    DONE = 3
    ERROR = 4


class CronjobEvent(enum.IntEnum):
    NONE = 0
    STOP = 1
    TIME_SYNC = 2


class Cronjob(threading.Thread):
    def __init__(self, name, cron_expression, actions):
        super().__init__()
        self.cron_expression = cron_expression
        self.name = name
        self.state = CronjobState.IDLE
        self._event = threading.Event()
        self._event_type = CronjobEvent.NONE
        self._event_lock = threading.RLock()

        if isinstance(actions, (list, dict)):
            self.actions = Procedure.build(
                name=name + '__Cron', _async=False, requests=actions
            )
        else:
            self.actions = actions

    def notify(self, event: CronjobEvent):
        with self._event_lock:
            self._event_type = event
            self._event.set()

    def run(self):
        set_thread_name(f'cron:{self.name}')
        self.wait()
        if self.should_stop():
            return

        self.state = CronjobState.RUNNING

        try:
            logger.info('Running cronjob {}'.format(self.name))
            context = {}

            if isinstance(self.actions, Procedure):
                response = self.actions.execute(_async=False, **context)
            else:
                response = self.actions(**context)

            logger.info('Response from cronjob {}: {}'.format(self.name, response))
            self.state = CronjobState.DONE
        except Exception as e:
            logger.exception(e)
            self.state = CronjobState.ERROR

    def wait(self):
        with self._event_lock:
            self.state = CronjobState.WAIT
            self._event.clear()
            self._event_type = CronjobEvent.TIME_SYNC

        while self._event_type == CronjobEvent.TIME_SYNC:
            now = get_now()
            self._event_type = CronjobEvent.NONE
            cron = croniter.croniter(self.cron_expression, now)
            next_run = cron.get_next(datetime.datetime)
            self._event.wait(max(0, (next_run - now).total_seconds()))

    def stop(self):
        self._event_type = CronjobEvent.STOP
        self._event.set()

    def should_stop(self):
        return self._event_type == CronjobEvent.STOP


class CronScheduler(threading.Thread):
    def __init__(self, jobs, poll_seconds: float = 0.5):
        super().__init__()
        self.jobs_config = jobs
        self._jobs = {}
        self._poll_seconds = max(1e-3, poll_seconds)
        self._should_stop = threading.Event()
        logger.info(
            'Cron scheduler initialized with {} jobs'.format(
                len(self.jobs_config.keys())
            )
        )

    def _get_job(self, name, config):
        job = self._jobs.get(name)
        if job and job.state not in [CronjobState.DONE, CronjobState.ERROR]:
            return job

        if isinstance(config, dict):
            self._jobs[name] = Cronjob(
                name=name,
                cron_expression=config['cron_expression'],
                actions=config['actions'],
            )
        elif is_functional_cron(config):
            self._jobs[name] = Cronjob(
                name=name, cron_expression=config.cron_expression, actions=config
            )
        else:
            raise AssertionError(
                'Expected type dict or function for cron {}, got {}'.format(
                    name, type(config)
                )
            )

        return self._jobs[name]

    def stop(self):
        for job in self._jobs.values():
            job.stop()
        self._should_stop.set()

    def should_stop(self):
        return self._should_stop.is_set()

    def run(self):
        logger.info('Running cron scheduler')

        while not self.should_stop():
            for (job_name, job_config) in self.jobs_config.items():
                job = self._get_job(name=job_name, config=job_config)
                if job.state == CronjobState.IDLE:
                    try:
                        job.start()
                    except Exception as e:
                        logger.warning(f'Could not start cronjob {job_name}: {e}')

            t_before_wait = get_now().timestamp()
            self._should_stop.wait(timeout=self._poll_seconds)
            t_after_wait = get_now().timestamp()
            time_drift = abs(t_after_wait - t_before_wait) - self._poll_seconds

            if not self.should_stop() and time_drift > 1:
                # If the system clock has been adjusted by more than one second
                # (e.g. because of DST change or NTP sync) then ensure that the
                # registered cronjobs are synchronized with the new datetime
                logger.info(
                    'System clock drift detected: %f secs. Synchronizing the cronjobs',
                    time_drift,
                )

                for job in self._jobs.values():
                    job.notify(CronjobEvent.TIME_SYNC)

        logger.info('Terminating cron scheduler')


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod
from typing import List, Optional

from platypush.plugins import Plugin, action


class LcdPlugin(Plugin, ABC):
    """
    Abstract class for plugins to communicate with LCD displays.

    Requires:

        * **RPLCD** (``pip install RPLCD``)
        * **RPi.GPIO** (``pip install RPi.GPIO``)

    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.lcd = None

    @staticmethod
    def _get_pin_mode(pin_mode: str) -> int:
        import RPi.GPIO
        pin_modes = ['BOARD', 'BCM']
        pin_mode = pin_mode.upper()
        assert pin_mode in pin_modes, 'Invalid pin_mode: {}. Supported modes: {}'.format(pin_mode, pin_modes)
        return getattr(RPi.GPIO, pin_mode).value

    @abstractmethod
    def _get_lcd(self):
        pass

    def _init_lcd(self):
        if self.lcd:
            return

        self.lcd = self._get_lcd()

    @action
    def close(self, clear: bool = False):
        """
        Close the handler to the LCD display and release the GPIO resources.

        :param clear: Clear the display as well on close (default: False).
        """
        if self.lcd:
            self.lcd.close(clear=clear)
            self.lcd = None

    @action
    def clear(self):
        """
        Clear the LCD display.
        """
        self._init_lcd()
        self.lcd.clear()

    @action
    def home(self):
        """
        Set cursor to initial position and reset any shifting.
        """
        self._init_lcd()
        self.lcd.home()

    @action
    def shift_display(self, amount: int):
        """
        Set cursor to initial position and reset any shifting.
        """
        self._init_lcd()
        self.lcd.shift_display(amount)

    @action
    def write_string(self, value: str, position: Optional[List[int]] = None):
        """
        Write a string to the display.

        :param value: String to be displayed.
        :param position: String position on the display as a 2-int list.
        """
        self._init_lcd()
        if position:
            self.lcd.cursor_pos = tuple(position)

        self.lcd.write_string(value)

    @action
    def set_cursor_pos(self, position: List[int]):
        """
        Change the position of the cursor on the display.

        :param position: New cursor position, as a list of two elements.
        """
        self._init_lcd()
        self.lcd.cursor_pos = tuple(position)

    @action
    def set_text_align(self, mode: str):
        """
        Change the text align mode.

        :param mode: Supported values: ``left``, ``right``.
        """
        modes = ['left', 'right']
        mode = mode.lower()
        assert mode in modes, 'Unsupported text mode: {}. Supported modes: {}'.format(
            mode, modes)

        self._init_lcd()
        self.lcd.text_align_mode = mode

    @action
    def enable_display(self):
        """
        Turn on the display.
        """
        self._init_lcd()
        self.lcd.display_enabled = True

    @action
    def disable_display(self):
        """
        Turn off the display.
        """
        self._init_lcd()
        self.lcd.display_enabled = False

    @action
    def toggle_display(self):
        """
        Toggle the display state.
        """
        self._init_lcd()
        self.lcd.display_enabled = not self.lcd.display_enabled

    @action
    def enable_backlight(self):
        """
        Enable the display backlight.
        """
        self._init_lcd()
        self.lcd.backlight_enabled = True

    @action
    def disable_backlight(self):
        """
        Disable the display backlight.
        """
        self._init_lcd()
        self.lcd.backlight_enabled = False

    @action
    def toggle_backlight(self):
        """
        Toggle the display backlight on/off.
        """
        self._init_lcd()
        self.lcd.backlight_enabled = not self.lcd.backlight_enabled

    @action
    def create_char(self, location: int, bitmap: List[int]):
        """
        Create a new character.
        The HD44780 supports up to 8 custom characters (location 0-7).

        :param location: The place in memory where the character is stored.
            Values need to be integers between 0 and 7.
        :param bitmap: The bitmap containing the character. This should be a
            list of 8 numbers, each representing a 5 pixel row.

        Example for the smiley character:

            .. code-block:: python

                [
                    0,   # 0b00000
                    10,  # 0b01010
                    10,  # 0b01010
                    0,   # 0b00000
                    17,  # 0b10001
                    17,  # 0b10001
                    14,  # 0b01110
                    0    # 0b00000
                ]

        """
        self._init_lcd()
        self.lcd.create_char(location=location, bitmap=tuple(bitmap))

    @action
    def command(self, value: int):
        """
        Send a raw command to the LCD.

        :param value: Command to be sent.
        """
        self._init_lcd()
        self.lcd.command(value)

    @action
    def write(self, value: int):
        """
        Write a raw byte to the LCD.

        :param value: Byte to be sent.
        """
        self._init_lcd()
        self.lcd.write(value)

    @action
    def cr(self):
        """
        Write a carriage return (``\\r``) character to the LCD.
        """
        self._init_lcd()
        self.lcd.cr()

    @action
    def lf(self):
        """
        Write a line feed (``\\n``) character to the LCD.
        """
        self._init_lcd()
        self.lcd.lf()

    @action
    def crlf(self):
        """
        Write a carriage return + line feed (``\\r\\n``) sequence to the LCD.
        """
        self._init_lcd()
        self.lcd.crlf()


# vim:sw=4:ts=4:et:



from datetime import datetime
from typing import Optional

from platypush.message.event import Event


class NewFeedEntryEvent(Event):
    """
    Event triggered when a new (RSS/Atom) feed entry is received.
    """

    def __init__(
            self, *, feed_url: str, url: str, title: Optional[str] = None, id: Optional[str] = None,
            feed_title: Optional[str] = None, published: Optional[datetime] = None, summary: Optional[str] = None,
            content: Optional[str] = None, **kwargs
    ):
        super().__init__(
            feed_url=feed_url, url=url, title=title, id=id, feed_title=feed_title,
            published=published, summary=summary, content=content, **kwargs
        )



import io
import json
import os
import threading

from typing import Any, Optional, IO

from platypush.config import Config
from platypush.message import Message
from platypush.plugins import Plugin, action


class MqttPlugin(Plugin):
    """
    This plugin allows you to send custom message to a message queue compatible
    with the MQTT protocol, see https://mqtt.org/

    Requires:

        * **paho-mqtt** (``pip install paho-mqtt``)

    """

    def __init__(
        self,
        host=None,
        port=1883,
        tls_cafile=None,
        tls_certfile=None,
        tls_keyfile=None,
        tls_version=None,
        tls_ciphers=None,
        tls_insecure=False,
        username=None,
        password=None,
        client_id=None,
        timeout=None,
        **kwargs,
    ):
        """
        :param host: If set, MQTT messages will by default routed to this host
            unless overridden in `send_message` (default: None)
        :type host: str

        :param port: If a default host is set, specify the listen port
            (default: 1883)
        :type port: int

        :param tls_cafile: If a default host is set and requires TLS/SSL,
            specify the certificate authority file (default: None)
        :type tls_cafile: str

        :param tls_certfile: If a default host is set and requires TLS/SSL,
            specify the certificate file (default: None)
        :type tls_certfile: str

        :param tls_keyfile: If a default host is set and requires TLS/SSL,
            specify the key file (default: None)
        :type tls_keyfile: str

        :param tls_version: If TLS/SSL is enabled on the MQTT server and it
            requires a certain TLS version, specify it here (default: None).
            Supported versions: ``tls`` (automatic), ``tlsv1``, ``tlsv1.1``,
            ``tlsv1.2``.
        :type tls_version: str

        :param tls_ciphers: If a default host is set and requires TLS/SSL,
            specify the supported ciphers (default: None)
        :type tls_ciphers: str

        :param tls_insecure: Set to True to ignore TLS insecure warnings
            (default: False).
        :type tls_insecure: bool

        :param username: If a default host is set and requires user
            authentication, specify the username ciphers (default: None)
        :type username: str

        :param password: If a default host is set and requires user
            authentication, specify the password ciphers (default: None)
        :type password: str

        :param client_id: ID used to identify the client on the MQTT server
            (default: None). If None is specified then
            ``Config.get('device_id')`` will be used.
        :type client_id: str

        :param timeout: Client timeout in seconds (default: None).
        :type timeout: int
        """

        super().__init__(**kwargs)

        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.client_id = client_id or Config.get('device_id')
        self.tls_cafile = self._expandpath(tls_cafile) if tls_cafile else None
        self.tls_certfile = self._expandpath(tls_certfile) if tls_certfile else None
        self.tls_keyfile = self._expandpath(tls_keyfile) if tls_keyfile else None
        self.tls_version = self.get_tls_version(tls_version)
        self.tls_insecure = tls_insecure
        self.tls_ciphers = tls_ciphers
        self.timeout = timeout

    @staticmethod
    def get_tls_version(version: Optional[str] = None):
        import ssl

        if not version:
            return None

        if isinstance(version, type(ssl.PROTOCOL_TLS)):
            return version

        if isinstance(version, str):
            version = version.lower()

        if version == 'tls':
            return ssl.PROTOCOL_TLS
        if version == 'tlsv1':
            return ssl.PROTOCOL_TLSv1
        if version == 'tlsv1.1':
            return ssl.PROTOCOL_TLSv1_1
        if version == 'tlsv1.2':
            return ssl.PROTOCOL_TLSv1_2

        assert 'Unrecognized TLS version: {}'.format(version)

    def _mqtt_args(self, **kwargs):
        return {
            'host': kwargs.get('host', self.host),
            'port': kwargs.get('port', self.port),
            'timeout': kwargs.get('timeout', self.timeout),
            'tls_certfile': kwargs.get('tls_certfile', self.tls_certfile),
            'tls_keyfile': kwargs.get('tls_keyfile', self.tls_keyfile),
            'tls_version': kwargs.get('tls_version', self.tls_version),
            'tls_ciphers': kwargs.get('tls_ciphers', self.tls_ciphers),
            'username': kwargs.get('username', self.username),
            'password': kwargs.get('password', self.password),
        }

    @staticmethod
    def _expandpath(path: Optional[str] = None) -> Optional[str]:
        return os.path.abspath(os.path.expanduser(path)) if path else None

    def _get_client(
        self,
        tls_cafile: Optional[str] = None,
        tls_certfile: Optional[str] = None,
        tls_keyfile: Optional[str] = None,
        tls_version: Optional[str] = None,
        tls_ciphers: Optional[str] = None,
        tls_insecure: Optional[bool] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ):
        from paho.mqtt.client import Client

        tls_cafile = self._expandpath(tls_cafile or self.tls_cafile)
        tls_certfile = self._expandpath(tls_certfile or self.tls_certfile)
        tls_keyfile = self._expandpath(tls_keyfile or self.tls_keyfile)
        tls_ciphers = tls_ciphers or self.tls_ciphers
        username = username or self.username
        password = password or self.password

        tls_version = tls_version or self.tls_version
        if tls_version:
            tls_version = self.get_tls_version(tls_version)
        if tls_insecure is None:
            tls_insecure = self.tls_insecure

        client = Client()

        if username and password:
            client.username_pw_set(username, password)
        if tls_cafile:
            client.tls_set(
                ca_certs=tls_cafile,
                certfile=tls_certfile,
                keyfile=tls_keyfile,
                tls_version=tls_version,
                ciphers=tls_ciphers,
            )

            client.tls_insecure_set(tls_insecure)

        return client

    @action
    def publish(
        self,
        topic: str,
        msg: Any,
        host: Optional[str] = None,
        port: Optional[int] = None,
        reply_topic: Optional[str] = None,
        timeout: int = 60,
        tls_cafile: Optional[str] = None,
        tls_certfile: Optional[str] = None,
        tls_keyfile: Optional[str] = None,
        tls_version: Optional[str] = None,
        tls_ciphers: Optional[str] = None,
        tls_insecure: Optional[bool] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        qos: int = 0,
    ):
        """
        Sends a message to a topic.

        :param topic: Topic/channel where the message will be delivered
        :param msg: Message to be sent. It can be a list, a dict, or a Message
            object.
        :param host: MQTT broker hostname/IP (default: default host configured
            on the plugin).
        :param port: MQTT broker port (default: default port configured on the
            plugin).
        :param reply_topic: If a ``reply_topic`` is specified, then the action
            will wait for a response on this topic.
        :param timeout: If ``reply_topic`` is set, use this parameter to
            specify the maximum amount of time to wait for a response (default:
            60 seconds).
        :param tls_cafile: If TLS/SSL is enabled on the MQTT server and the
            certificate requires a certificate authority to authenticate it,
            `ssl_cafile` will point to the provided ca.crt file (default:
            None).
        :param tls_certfile: If TLS/SSL is enabled on the MQTT server and a
            client certificate it required, specify it here (default: None).
        :param tls_keyfile: If TLS/SSL is enabled on the MQTT server and a
            client certificate key it required, specify it here (default:
            None).
        :param tls_version: If TLS/SSL is enabled on the MQTT server and it
            requires a certain TLS version, specify it here (default: None).
            Supported versions: ``tls`` (automatic), ``tlsv1``, ``tlsv1.1``,
            ``tlsv1.2``.
        :param tls_insecure: Set to True to ignore TLS insecure warnings
            (default: False).
        :param tls_ciphers: If TLS/SSL is enabled on the MQTT server and an
            explicit list of supported ciphers is required, specify it here
            (default: None).
        :param username: Specify it if the MQTT server requires authentication
            (default: None).
        :param password: Specify it if the MQTT server requires authentication
            (default: None).
        :param qos: Quality of Service (_QoS_) for the message - see `MQTT QoS
            <https://assetwolf.com/learn/mqtt-qos-understanding-quality-of-service>`_
            (default: 0).
        """
        response_buffer = io.BytesIO()
        client = None

        try:
            # Try to parse it as a platypush message or dump it to JSON from a dict/list
            if isinstance(msg, (dict, list)):
                msg = json.dumps(msg)

                try:
                    msg = Message.build(json.loads(msg))
                except Exception as e:
                    self.logger.debug(f'Not a valid JSON: {str(e)}')

            host = host or self.host
            port = port or self.port or 1883
            assert host, 'No host specified'

            client = self._get_client(
                tls_cafile=tls_cafile,
                tls_certfile=tls_certfile,
                tls_keyfile=tls_keyfile,
                tls_version=tls_version,
                tls_ciphers=tls_ciphers,
                tls_insecure=tls_insecure,
                username=username,
                password=password,
            )

            client.connect(host, port, keepalive=timeout)
            response_received = threading.Event()

            if reply_topic:
                client.on_message = self._response_callback(
                    reply_topic=reply_topic,
                    event=response_received,
                    buffer=response_buffer,
                )
                client.subscribe(reply_topic)

            client.publish(topic, str(msg), qos=qos)
            if not reply_topic:
                return

            client.loop_start()
            ok = response_received.wait(timeout=timeout)
            if not ok:
                raise TimeoutError('Response timed out')
            return response_buffer.getvalue()
        finally:
            response_buffer.close()

            if client:
                try:
                    client.loop_stop()
                except Exception as e:
                    self.logger.warning(f'Could not stop client loop: {e}')

                client.disconnect()

    @staticmethod
    def _response_callback(reply_topic: str, event: threading.Event, buffer: IO[bytes]):
        def on_message(client, _, msg):
            if msg.topic != reply_topic:
                return

            buffer.write(msg.payload)
            client.loop_stop()
            event.set()

        return on_message

    @action
    def send_message(self, *args, **kwargs):
        """
        Alias for :meth:`platypush.plugins.mqtt.MqttPlugin.publish`.
        """
        return self.publish(*args, **kwargs)


# vim:sw=4:ts=4:et:
